from typing import Dict, Union
from depccg.cat import Category, Atom, Feature


class Unification(object):
    """This performs unification on category variables.
    Usage:
    >>> uni = Unification("a/b", "b")
    >>> x = Category.parse("S[X]/NP[X]")
    >>> y = Category.parse("NP[mod]")
    >>> uni(x, y)
    True
    >>> uni["a"]
    S[mod]

    Args:
        meta_x: a string pattern (e.g., "a/b") to match against the first argument (x above).
        meta_y: a string pattern ("b") to match against the first argument (y above).
    """

    def __init__(
        self,
        meta_x: Union[str, Category],
        meta_y: Union[str, Category],
    ) -> None:
        self.cats: Dict[str, Category] = {}
        # mapping of variable feature to its instantiation
        self.mapping: Dict[Feature, Feature] = {}
        self.meta_x = (
            Category.parse(meta_x) if isinstance(meta_x, str) else meta_x
        )
        self.meta_y = (
            Category.parse(meta_y) if isinstance(meta_y, str) else meta_y
        )
        # meta variables to feature values
        self.x_features: Dict[str, Feature] = {}
        self.y_features: Dict[str, Feature] = {}
        self.success = False
        self.done = False

    def __call__(self, x: Category, y: Category) -> bool:
        if self.done:
            raise RuntimeError(
                "cannot use the same Unification object more than once."
            )
        self.done = True

        def scan_deep(s: Category, v: str, index: int, results: Dict[str, Feature]):
            if s.is_functor:
                index = scan_deep(s.left, v, index, results)
                index = scan_deep(s.right, v, index, results)
                return index
            results[f'{v}{index}'] = s.feature
            return index + 1

        def scan(s: Category, t: Category, results: Dict[str, Feature]) -> bool:
            # collect categories corresponding to meta variables
            if s.is_atomic:
                if s.base in self.cats and not (t ^ self.cats[s.base]):
                    return False
                self.cats[s.base] = t

            if (
                s.is_functor and t.is_functor and (
                    s.slash == t.slash or '|' in (s.slash, t.slash))
            ):
                return (
                    scan(s.left, t.left, results) and scan(
                        s.right, t.right, results)
                )
            elif s.is_atomic and t.is_functor:
                scan_deep(t, s.base, 0, results)
                return True
            elif s.is_atomic and t.is_atomic:
                if t.feature is not None:
                    results[s.base] = t.feature
                return True
            return False  # s.is_functor and t.is_atomic

        self.success = (
            scan(self.meta_x, x, self.x_features) and scan(
                self.meta_y, y, self.y_features)
        )

        if not self.success:
            return False

        # meta variables such as a, b, c, d, etc.
        meta_vars = set(self.x_features.keys()) & set(self.y_features.keys())

        if len(meta_vars) == 0:
            return True

        for var in meta_vars:
            x_feature = self.x_features[var]
            y_feature = self.y_features[var]

            # these pairs can match: (NP[nb], NP[conj]), (NP, NP[conj]), (NP[X], NP[conj])
            if x_feature.unifies(y_feature):
                # if (NP[X], NP[conj]), further memorize the matching `X := conj`
                # if there're two of `X := conj` and `X := nb`, choose one arbitrarily
                if x_feature.is_variable:
                    self.mapping[x_feature] = y_feature

            elif y_feature.unifies(x_feature):
                if y_feature.is_variable:
                    self.mapping[y_feature] = x_feature
            else:
                self.success = False
                return False

        return True

    def __getitem__(self, key: str) -> Category:

        def rec(x: Category) -> Category:
            if x.is_functor:
                return x.functor(rec(x.left), rec(x.right))
            else:
                if x.feature in self.mapping:
                    return Atom(x.base, self.mapping[x.feature])
                else:
                    return x

        assert self.success, \
            ("the unification has not been successful. "
             "Unification.__getitem__ is not callable in that case.")

        if key not in self.cats:
            raise KeyError(f'meta category `{key}` has not been observed.')
        return rec(self.cats[key])




from io import StringIO
from depccg.tree import Tree


def deriv_of(tree: Tree) -> str:
    """ascii art-like derivation tree string.

    Args:
        tree (Tree): tree object

    Returns:
        str: derivation tree string
    """

    catstr = ''
    wordstr = ''
    for leaf in tree.leaves:
        str_cat = str(leaf.cat)
        str_word = leaf.word
        nextlen = 2 + max(len(str_word), len(str_cat))
        lcatlen = (nextlen - len(str_cat)) // 2
        rcatlen = lcatlen + (nextlen - len(str_cat)) % 2
        catstr += ' ' * lcatlen + str_cat + ' ' * rcatlen
        lwordlen = (nextlen - len(str_word)) // 2
        rwordlen = lwordlen + (nextlen - len(str_word)) % 2
        wordstr += ' ' * lwordlen + str_word + ' ' * rwordlen

    def rec(lwidth, node):
        rwidth = lwidth

        if node.is_leaf:
            return max(
                rwidth,
                2 + lwidth + len(str(node.cat)),
                2 + lwidth + len(node.word)
            )

        else:
            for child in node.children:
                rwidth = max(rwidth, rec(rwidth, child))

            print(
                lwidth * ' ' + (rwidth - lwidth) * '-' + str(node.op_symbol),
                file=output
            )

            result = str(node.cat)
            pad_len = (rwidth - lwidth - len(result)) // 2 + lwidth
            print(pad_len * ' ' + result, file=output)

            return rwidth

    with StringIO() as output:
        print(catstr.rstrip(), file=output)
        print(wordstr.rstrip(), file=output)
        rec(0, tree)
        return output.getvalue()



from typing import List
from cython.operator cimport dereference as deref
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.unordered_set cimport unordered_set
from libcpp.unordered_map cimport unordered_map
cimport numpy as np
import copy

from tqdm import tqdm
from depccg.tree import Tree, ScoredTree
from depccg.cat import Category
from depccg.types import ScoringResult

cdef extern from "<limits>":
    cdef unsigned UINT_MAX


cdef extern from "depccg/parsing.h" namespace "parsing":
    cdef struct cell_item:
        bint fin
        unsigned cat
        cell_item *left
        cell_item *right
        float in_score
        float out_score
        unsigned start_of_span
        unsigned span_length
        unsigned head_id
        unsigned rule_id

        float score()


cdef extern from "depccg/parsing.h":
    cdef struct combinator_result:
        unsigned cat_id
        unsigned rule_id
        bint head_is_left
        string op_string
        string op_symbol

    ctypedef unordered_map[pair[unsigned, unsigned], vector[combinator_result]] cache_type

    ctypedef int (*scaffold_type)(void *callback_func, unsigned x, unsigned y, vector[combinator_result] *results) except -1

    ctypedef unsigned (*finalizer_type)(cell_item *, unsigned *, cache_type *cache, void *)

    cdef struct config:
        unsigned num_tags
        float unary_penalty
        float beta
        bint use_beta
        unsigned pruning_size
        unsigned nbest
        unsigned max_step

    cdef unsigned parse_sentence(
        float *tag_scores,
        float *dep_scores,
        unsigned length,
        const unordered_set[unsigned] &possible_root_cats,
        void *binary_callback,
        void *unary_callback,
        finalizer_type finalizer_callback,
        scaffold_type scaffold,
        void *finalizer_args,
        cache_type *cache,
        config *config) except +


cdef int scaffold(
    void *callback_func,
    unsigned x,
    unsigned y,
    vector[combinator_result] *results,
) except -1:
    cdef list py_results
    cdef cat_id, rule_id
    cdef bint head_is_left
    cdef combinator_result c_result

    py_results = (<object>callback_func)(<object>x, <object>y)
    for cat_id, rule_id, result in py_results:
        c_result.cat_id = cat_id
        c_result.rule_id = rule_id
        c_result.head_is_left = result.head_is_left
        c_result.op_string = result.op_string.encode('utf-8')
        c_result.op_symbol = result.op_symbol.encode('utf-8')
        results.push_back(c_result)
    return 0;


cdef init_config(config *c_config, dict kwargs):
    c_config.num_tags = kwargs['num_tags']
    c_config.unary_penalty = kwargs.pop('unary_penalty', 0.1)
    c_config.beta = kwargs.pop('beta', 0.00001)
    c_config.use_beta = kwargs.pop('use_beta', True)
    c_config.pruning_size = kwargs.pop('pruning_size', 50)
    c_config.nbest = kwargs.pop('nbest', 1)
    c_config.max_step = kwargs.pop('max_step', 10000000)


cdef unsigned retrieve_tree(
    cell_item *item,
    unsigned *token_id,
    cache_type *cache,
    void *c_kwargs,
):
    """
    retrieve (n-best) parsing results from cell_item.
    they are placed in kwargs['stack']
    """
    cdef pair[unsigned, unsigned] key
    cdef object kwargs = <object>c_kwargs

    if item.fin:
        kwargs['scores'].append(item.score())
        return retrieve_tree(item.left, token_id, cache, c_kwargs)

    cat = kwargs['categories'][item.cat]
    stack = kwargs['stack']
    if item.left == NULL and item.right == NULL:
        stack.append(
            Tree.make_terminal(kwargs['tokens'][token_id[0]], cat)
        )
        token_id[0] += 1
        return item.cat
    else:
        child_cat = retrieve_tree(item.left, token_id, cache, c_kwargs)
        kwargs = <object>c_kwargs
        if item.right == NULL:
            child = stack.pop()
            key.first = child_cat
            key.second = -1
            combinator_result = cache[0][key][item.rule_id]
            stack.append(
                Tree.make_unary(
                    cat,
                    child,
                    combinator_result.op_string.decode('utf-8'),
                    combinator_result.op_symbol.decode('utf-8'),
                )
            )
            return item.cat
        right_child_cat = retrieve_tree(item.right, token_id, cache, c_kwargs)
        key.first = child_cat
        key.second = right_child_cat
        kwargs = <object>c_kwargs
        right = stack.pop()
        left = stack.pop()
        combinator_result = cache[0][key][item.rule_id]
        stack.append(
            Tree.make_binary(
                cat,
                left,
                right,
                combinator_result.op_string.decode('utf-8'),
                combinator_result.op_symbol.decode('utf-8'),
            )
        )
        return item.cat


def run(
    list doc,
    list scoring_results,
    list categories,
    apply_binary_rules,
    apply_unary_rules,
    object possible_root_cats,
    process_id=0,
    **kwargs
) -> List[ScoredTree]:

    def failed():
        return [
            ScoredTree(
                tree=Tree.make_terminal("FAILED", Category.parse("NP")),
                score=-float('inf')
            )
        ]

    if len(set(categories)) != len(categories):
        raise RuntimeError(
            'argument `categories` cannot contain duplicate elements.'
        )

    categories_ = copy.copy(categories)

    category_ids = {
        category: index
        for index, category in enumerate(categories_)
    }

    def maybe_add_and_get(cat):
        if cat not in category_ids:
            categories_.append(cat)
            category_ids[cat] = len(category_ids)
        if len(categories_) >= UINT_MAX:
            raise RuntimeError('too many categories')
        return category_ids[cat]

    def binary_callback(x_id, y_id):
        x, y = categories_[x_id], categories_[y_id]

        results = []
        for rule_id, result in enumerate(apply_binary_rules(x, y)):
            cat_id = maybe_add_and_get(result.cat)
            results.append((cat_id, rule_id, result))
        return results

    def unary_callback(x_id, _):
        x = categories_[x_id]

        results = []
        for rule_id, result in enumerate(apply_unary_rules(x)):
            cat_id = maybe_add_and_get(result.cat)
            results.append((cat_id, rule_id, result))
        return results

    cdef list tokens
    cdef np.ndarray[float, ndim=2, mode='c'] tag_scores
    cdef np.ndarray[float, ndim=2, mode='c'] dep_scores
    cdef float *c_tag_scores
    cdef float *c_dep_scores
    cdef unsigned length, status
    cdef unordered_set[unsigned] c_possible_root_cat
    cdef cache_type c_cache
    cdef config c_config
    init_config(&c_config, kwargs)

    cdef unsigned cat_id
    for cat in possible_root_cats:
        cat_id = maybe_add_and_get(cat)
        c_possible_root_cat.insert(cat_id)

    all_results = []
    iter_ = tqdm(
        list(zip(doc, scoring_results)),
        desc=f'#{process_id:>2} ',
        position=process_id
    )
    for tokens, (tag_scores, dep_scores) in iter_:
        c_tag_scores = <float*>tag_scores.data
        c_dep_scores = <float*>dep_scores.data
        length = len(tokens)

        if (
            'max_length' in kwargs
            and len(tokens) > kwargs['max_length']
        ):
            all_results.append(failed())
            continue

        results = []
        scores = []
        finalizer_args = {
            'categories': categories_,
            'tokens': tokens,
            'stack': results,
            'scores': scores,
        }

        status = parse_sentence(
            c_tag_scores,
            c_dep_scores,
            length,
            c_possible_root_cat,
            <void*>binary_callback,
            <void*>unary_callback,
            retrieve_tree,
            scaffold,
            <void*>finalizer_args,
            &c_cache,
            &c_config,
        )

        if status > 0:
            all_results.append(failed())
            continue

        if len(results) != len(scores):
            raise RuntimeError(
                'unexpected behavior occured during parsing'
            )

        all_results.append(
            [
                ScoredTree(tree=tree, score=score)
                for tree, score in zip(results, scores)
            ]
        )

    return all_results



import argparse
from depccg.instance_models import download, AVAILABLE_MODEL_VARIANTS
from depccg.annotator import (
    english_annotator, japanese_annotator
)


def add_common_parser_arguments(parser, main_fun):
    parser.add_argument(
        '-c',
        '--config',
        help='json config file specifying the set of unary rules used, etc.')
    parser.add_argument(
        '-m',
        '--model',
        help='path to model directory')
    parser.add_argument(
        '-p',
        '--num-processes',
        default=4,
        type=int,
        help='number of processes used for parsing')
    parser.add_argument(
        '-i',
        '--input',
        default=None,
        help='a file with tokenized sentences in each line')
    # parser.add_argument(
    #     '-w',
    #     '--weights',
    #     default=None,
    #     help='a file that contains weights (p_tag, p_dep)')
    parser.add_argument(
        '--gpu',
        type=int,
        default=-1,
        help='specify gpu id')
    parser.add_argument(
        '--batchsize',
        type=int,
        default=32,
        help='batchsize in supertagger')
    parser.add_argument(
        '--nbest',
        type=int,
        default=1,
        help='output N best parses')
    parser.add_argument(
        '-I',
        '--input-format',
        default='raw',
        choices=['raw', 'POSandNERtagged'],
        # choices=['raw', 'POSandNERtagged', 'json', 'partial'],
        help='input format')
    parser.add_argument(
        '--unary-penalty',
        default=0.1,
        type=float,
        help='penalty to use a unary rule')
    parser.add_argument(
        '--beta',
        default=0.00001,
        type=float,
        help='parameter used to filter categories with lower probabilities')
    parser.add_argument(
        '--pruning-size',
        default=50,
        type=int,
        help='use only the most probable supertags per word')
    parser.add_argument(
        '--disable-beta',
        action='store_true',
        help='disable the use of the beta value')
    parser.add_argument(
        '--disable-category-dictionary',
        action='store_true',
        help=('disable a category dictionary that maps'
              ' words to most likely supertags'))
    parser.add_argument(
        '--disable-seen-rules',
        action='store_true',
        help='')
    parser.add_argument(
        '--max-length',
        default=250,
        type=int,
        help=('give up parsing a sentence that contains'
              ' more words than this value'))
    parser.add_argument(
        '--max-step',
        default=10000000,
        type=int,
        help=('give up parsing when the number of times'
              ' of popping agenda items exceeds this value'))
    parser.add_argument(
        '--semantic-templates',
        help='semantic templates used in "ccg2lambda" format output')
    parser.add_argument(
        '--silent',
        action='store_true')
    parser.set_defaults(func=main_fun)

    subparsers = parser.add_subparsers()
    download_parser = subparsers.add_parser('download')
    download_parser.add_argument(
        'VARIANT',
        nargs='?',
        default=None,
        choices=AVAILABLE_MODEL_VARIANTS[parser.get_default('lang')])

    download_parser.set_defaults(
        func=lambda args: download(args.lang, args.VARIANT)
    )


def parse_args(main_fun):
    parser = argparse.ArgumentParser('depccg')
    parser.set_defaults(func=lambda _: parser.print_help())
    subparsers = parser.add_subparsers()

    english_parser = subparsers.add_parser('en')
    english_parser.set_defaults(lang='en')
    add_common_parser_arguments(english_parser, main_fun)
    english_parser.add_argument(
        '-a',
        '--annotator',
        default=None,
        help='annotate POS, named entity, and lemmas using this library',
        choices=english_annotator.keys())
    english_parser.add_argument(
        '-f',
        '--format',
        default='auto',
        choices=[
            'auto', 'auto_extended', 'deriv', 'xml',
            'conll', 'html', 'prolog', 'jigg_xml', 'ptb',
            'ccg2lambda', 'jigg_xml_ccg2lambda', 'json'
        ],
        help='output format')
    english_parser.add_argument(
        '--root-cats',
        default='S[dcl]|S[wq]|S[q]|S[qem]|NP',
        help=('"|" separated list of categories '
              'allowed to be at the root of a tree.')
    ),
    english_parser.add_argument(
        '--tokenize',
        action='store_true',
        help='tokenize input sentences')

    japanese_parser = subparsers.add_parser('ja')
    japanese_parser.set_defaults(lang='ja')
    add_common_parser_arguments(japanese_parser, main_fun)
    japanese_parser.add_argument(
        '-a',
        '--annotator',
        default='janome',
        help=('annotate POS, named entity,'
              ' and lemmas using this library'),
        choices=japanese_annotator.keys())
    japanese_parser.add_argument(
        '-f',
        '--format',
        default='ja',
        choices=[
            'auto', 'deriv', 'ja', 'conll',
            'html', 'jigg_xml', 'ptb', 'ccg2lambda',
            'jigg_xml_ccg2lambda', 'json', 'prolog'
        ],
        help='output format')
    japanese_parser.add_argument(
        '--root-cats',
        default=(
            'NP[case=nc,mod=nm,fin=f]|'
            'NP[case=nc,mod=nm,fin=t]|'
            'S[mod=nm,form=attr,fin=t]|'
            'S[mod=nm,form=base,fin=f]|'
            'S[mod=nm,form=base,fin=t]|'
            'S[mod=nm,form=cont,fin=f]|'
            'S[mod=nm,form=cont,fin=t]|'
            'S[mod=nm,form=da,fin=f]|'
            'S[mod=nm,form=da,fin=t]|'
            'S[mod=nm,form=hyp,fin=t]|'
            'S[mod=nm,form=imp,fin=f]|'
            'S[mod=nm,form=imp,fin=t]|'
            'S[mod=nm,form=r,fin=t]|'
            'S[mod=nm,form=s,fin=t]|'
            'S[mod=nm,form=stem,fin=f]|'
            'S[mod=nm,form=stem,fin=t]'

        ),
        help=('"|" separated list of categories '
              'allowed to be at the root of a tree.')
    )
    japanese_parser.add_argument(
        '--pre-tokenized',
        dest='tokenize',
        action='store_false',
        help=('the input is pre-tokenized'
              ' (for running parsing experiments etc.)'))

    args = parser.parse_args()
    args.func(args)



from typing import Dict, Any, List, Optional

import numpy
from allennlp.common.util import JsonDict
from allennlp.data import DatasetReader
from allennlp.models import Model
from allennlp.predictors.predictor import Predictor

import depccg.parsing
from depccg.types import ScoringResult, Token
from depccg.allennlp.predictor.supertagger_predictor import SupertaggerPredictor
from depccg.allennlp.utils import read_params
from depccg.cat import Category
from depccg.printer.my_json import json_of


@Predictor.register('parser-predictor')
class ParserPredictor(SupertaggerPredictor):

    def __init__(
        self,
        model: Model,
        dataset_reader: DatasetReader,
        grammar_json_path: str,
        parsing_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(model, dataset_reader)
        (
            self.apply_binary_rules,
            self.apply_unary_rules,
            self.category_dict,
            self.root_categories,
        ) = read_params(grammar_json_path)
        self.parsing_kwargs = parsing_kwargs or {}

    def _make_json(self, output_dicts: List[Dict[str, Any]]) -> List[JsonDict]:

        categories = None
        score_results = []
        doc = []
        for output_dict in super()._make_json(output_dicts):
            if categories is None:
                categories = [
                    Category.parse(category)
                    for category in output_dict['categories']
                ]

            tokens = [
                Token.of_word(word)
                for word in output_dict['words'].split(' ')
            ]
            doc.append(tokens)

            dep_scores = numpy.array(output_dict['heads']) \
                .reshape(output_dict['heads_shape']) \
                .astype(numpy.float32)
            tag_scores = numpy.array(output_dict['head_tags']) \
                .reshape(output_dict['head_tags_shape']) \
                .astype(numpy.float32)
            score_results.append(ScoringResult(tag_scores, dep_scores))

        if self.category_dict is not None:
            doc, score_results = depccg.parsing.apply_category_filters(
                doc,
                score_results,
                categories,
                self.category_dict,
            )

        results = depccg.parsing.run(
            doc,
            score_results,
            categories,
            self.root_categories,
            self.apply_binary_rules,
            self.apply_unary_rules,
            **self.parsing_kwargs,
        )

        for output_dict, trees in zip(output_dicts, results):
            output_dict['trees'] = []
            for tree, log_prob in trees:
                tree_dict = json_of(tree)
                tree_dict['log_prob'] = log_prob
                output_dict['trees'].append(tree_dict)

        return output_dicts



from typing import List, Union, Optional
import html
import re
from depccg.tree import Tree, ScoredTree


_MATHML_SUBTREE_NONTERMINAL = '''\
<mrow>
  <mfrac {3} linethickness='2px'>
    <mrow>{0}</mrow>
    <mstyle mathcolor='Red'>{1}</mstyle>
  </mfrac>
  <mtext mathsize='0.8' mathcolor='Black'>{2}</mtext>
</mrow>
'''


_MATHML_SUBTREE_TERMINAL = '''\
<mrow>
  <mfrac linethickness='2px'>
    <mtext mathsize='1.0' mathcolor='Black'>{0}</mtext>
    <mstyle mathcolor='Red'>{1}</mstyle>
  </mfrac>
  <mtext mathsize='0.8' mathcolor='Black'>lex</mtext>
</mrow>
'''


_MATHML_MAIN = '''\
<!doctype html>
<html lang='en'>
<head>
  <meta charset='UTF-8'>
  <style>
    body {{
      font-size: 1em;
    }}
  </style>
  <script type="text/javascript"
     src="http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML">
  </script>
</head>
<body>
  {0}
</body>
</html>
'''


def _mathml_subtree(tree: Tree, bgcolor: Optional[str] = None) -> str:
    bgcolor = f'mathbackground={bgcolor}' if bgcolor else ''
    cat_str = _mathml_cat(str(tree.cat))
    if not tree.is_leaf:
        children_str = ''.join(map(_mathml_subtree, tree.children))
        return _MATHML_SUBTREE_NONTERMINAL.format(
            children_str, cat_str, html.escape(tree.op_string), bgcolor
        )
    else:
        return _MATHML_SUBTREE_TERMINAL.format(html.escape(tree.word), cat_str)


def _mathml_cat(cat: str) -> str:
    cats_feats = re.findall(r'([\w\\/()]+)(\[.+?\])*', cat)
    mathml_str = ''
    for cat, feat in cats_feats:
        cat = html.escape(cat)
        feat = html.escape(feat)
        cat_mathml = f'''\
<mi mathvariant='italic'
  mathsize='1.0' mathcolor='Red'>{cat}</mi>'''
        if feat != '':
            mathml_str += f'''\
<msub>{cat_mathml}
  <mrow>
  <mi mathvariant='italic'
    mathsize='0.8' mathcolor='Purple'>{feat}</mi>
  </mrow>
</msub>'''
        else:
            mathml_str += cat_mathml
    return mathml_str


def to_mathml(nbest_trees: List[List[Union[Tree, ScoredTree]]]) -> str:
    """convert parsing results to a html string containing all the trees.

    Args:
        nbest_trees (List[List[Union[Tree, ScoredTree]]]): parsing results

    Returns:
        str: html string
    """

    result = ''
    for sentence_index, trees in enumerate(nbest_trees, 1):
        if isinstance(trees[0], ScoredTree):
            words = trees[0].tree.word
        else:
            words = trees[0].word

        result += f'<p>ID={sentence_index}: {words}</p>'

        for tree in trees:
            if isinstance(tree, ScoredTree):
                tree, prob = tree
                result += f'<p>Log prob={prob:.5e}</p>'

            tree_str = tree if isinstance(tree, str) else _mathml_subtree(tree)
            result += f'<math xmlns="http://www.w3.org/1998/Math/MathML">{tree_str}</math>'

    return _MATHML_MAIN.format(result)



from typing import Tuple, List, Iterator, NamedTuple
from depccg.tree import Tree
from depccg.cat import Category
from depccg.lang import get_global_language
from depccg.types import Token
from depccg.grammar import guess_combinator_by_triplet
from depccg.grammar import en, ja
from lxml import etree
import logging

logger = logging.getLogger(__name__)


BINARY_RULES = {
    'en': en.apply_binary_rules,
    'ja': ja.apply_binary_rules,
}


class ReaderResult(NamedTuple):
    name: str
    tokens: List[Token]
    tree: Tree


class _AutoLineReader(object):
    def __init__(self, line):
        self.line = line
        self.index = 0
        self.word_id = -1
        self.binary_rules = BINARY_RULES[get_global_language()]
        self.tokens = []

    def next(self):
        end = self.line.find(' ', self.index)
        res = self.line[self.index:end]
        self.index = end + 1
        return res

    def check(self, text, offset=0):
        if self.line[self.index + offset] != text:
            raise RuntimeError(f'failed to parse: {self.line}')

    def peek(self):
        return self.line[self.index]

    def parse(self):
        tree = self.next_node()
        return tree, self.tokens

    @property
    def next_node(self):
        if self.line[self.index + 2] == 'L':
            return self.parse_leaf
        elif self.line[self.index + 2] == 'T':
            return self.parse_tree
        else:
            raise RuntimeError(f'failed to parse: {self.line}')

    def parse_leaf(self):
        self.word_id += 1
        self.check('(')
        self.check('<', 1)
        self.check('L', 2)
        self.next()
        cat = Category.parse(self.next())
        tag1 = self.next()  # modified POS tag
        tag2 = self.next()  # original POS
        word = self.next().replace('\\', '')
        token = Token(
            word=word,
            pos=tag1,
            tag1=tag1,
            tag2=tag2
        )
        self.tokens.append(token)
        if word == '-LRB-':
            word = "("
        elif word == '-RRB-':
            word = ')'
        self.next()
        return Tree.make_terminal(token, cat)

    def parse_tree(self):
        self.check('(')
        self.check('<', 1)
        self.check('T', 2)
        self.next()
        cat = Category.parse(self.next())
        head_is_left = self.next() == '0'
        self.next()
        children = []
        while self.peek() != ')':
            children.append(self.next_node())
        self.next()
        if len(children) == 2:
            left, right = children
            rule = guess_combinator_by_triplet(
                self.binary_rules, cat, left.cat, right.cat
            )
            return Tree.make_binary(
                cat, left, right, rule.op_string, rule.op_symbol, head_is_left
            )
        elif len(children) == 1:
            return Tree.make_unary(cat, children[0])
        else:
            raise RuntimeError(f'failed to parse: {self.line}')


def read_auto(filename: str) -> Iterator[ReaderResult]:
    """read traditional AUTO file used for CCGBank
    English CCGbank contains some unwanted categories such as (S\\NP)\\(S\\NP)[conj].
    This reads the treebank while taking care of those categories.

    Args:
        filename (str): file name string

    Yields:
        Iterator[ReaderResult]: iterator object containing parse results
    """

    for line in open(filename):
        line = line.strip()
        if len(line) == 0:
            continue
        if line.startswith("ID"):
            name = line
        else:
            tokens = []
            for token in line.split(' '):
                if token.endswith(')[conj]'):
                    token = token[:-6]
                tokens.append(token)
            line = ' '.join(tokens)
            tree, tokens = _AutoLineReader(line).parse()
            yield ReaderResult(name, tokens, tree)


def read_xml(filename: str) -> Iterator[ReaderResult]:
    """read XML format file commonly used by C&C.

    Args:
        filename (str): file name string

    Yields:
        Iterator[ReaderResult]: iterator object containing parse results
    """

    binary_rules = BINARY_RULES[get_global_language()]

    def parse(tree):
        def rec(node):
            attrib = node.attrib
            if node.tag == 'rule':
                cat = Category.parse(attrib['cat'])
                children = [rec(child) for child in node.getchildren()]
                if len(children) == 1:
                    return Tree.make_unary(cat, children[0])
                else:
                    assert len(children) == 2
                    left, right = children
                    rule = guess_combinator_by_triplet(
                        binary_rules, cat, left.cat, right.cat
                    )
                    return Tree.make_binary(
                        cat, left, right, rule.op_string, rule.op_symbol, rule.head_is_left
                    )
            else:
                assert node.tag == 'lf'
                cat = Category.parse(attrib['cat'])
                token = Token(
                    word=attrib['word'],
                    pos=attrib['pos'],
                    entity=attrib['entity'],
                    lemma=attrib['lemma'],
                    chunk=attrib['chunk']
                )
                tokens.append(token)
                return Tree.make_terminal(token, cat)
        tokens = []
        tree = rec(tree)
        return tokens, tree

    trees = etree.parse(filename).getroot().xpath('ccg')
    for tree in trees:
        name = '_'.join(f'{k}={v}' for k, v in tree.items())
        yield ReaderResult(name, *parse(tree[0]))


def read_jigg_xml(filename: str) -> Iterator[ReaderResult]:
    """read XML format file used by Jigg.

    Args:
        filename (str): file name string

    Yields:
        Iterator[ReaderResult]: iterator object containing parse results
    """

    binary_rules = BINARY_RULES[get_global_language()]
    # TODO

    def try_get_surface(token):
        if 'word' in token:
            return token.word
        elif 'surf' in token:
            return token.surf
        else:
            raise RuntimeError(
                'the attribute for the token\'s surface form is unknown'
            )

    def parse(tree, tokens):
        def rec(node):
            attrib = node.attrib
            if 'terminal' not in attrib:
                cat = Category.parse(attrib['category'])
                children = [
                    rec(spans[child])
                    for child in attrib['child'].split(' ')
                ]
                if len(children) == 1:
                    return Tree.make_unary(cat, children[0])
                else:
                    assert len(children) == 2
                    left, right = children
                    rule = guess_combinator_by_triplet(
                        binary_rules, cat, left.cat, right.cat
                    )
                    return Tree.make_binary(
                        cat, left, right, rule.op_string, rule.op_symbol, rule.head_is_left
                    )
            else:
                cat = Category.parse(attrib['category'])
                word = try_get_surface(tokens[attrib['terminal']])
                return Tree.make_terminal(word, cat)

        spans = {span.attrib['id']: span for span in tree.xpath('./span')}
        return rec(spans[tree.attrib['root']])

    trees = etree.parse(filename).getroot()
    sentences = trees[0][0].xpath('sentence')
    for sentence in sentences:
        token_and_ids = []
        for token in sentence.xpath('.//token'):
            token_attribs = dict(token.attrib)
            token_id = token_attribs['id']
            for no_need in ['id', 'start', 'cat']:
                if no_need in token_attribs:
                    del token_attribs[no_need]
            token_and_ids.append((token_id, Token(**token_attribs)))

        tokens = [token for _, token in token_and_ids]

        for ccg in sentence.xpath('./ccg'):
            tree = parse(ccg, dict(token_and_ids))
            yield ReaderResult(ccg.attrib['id'], tokens, tree)


def _parse_ptb(tree_string: str) -> Tuple[Tree, List[Token]]:
    """parse a S-expression like PTB-format tree

    Args:
        tree_string (str): S-expression

    Raises:
        RuntimeError: when parsing fails.

    Returns:
        Tuple[Tree, List[Token]]: Tree object and tokens
    """
    binary_rules = BINARY_RULES[get_global_language()]
    assert tree_string.startswith('(ROOT ')
    buf = list(reversed(tree_string[6:-1].split(' ')))
    stack = []
    tokens = []
    position = 0

    def reduce(item: str) -> None:
        nonlocal position
        if item[-1] != ')':
            token = Token(word=item)
            tokens.append(token)
            stack.append(item)
            return

        reduce(item[:-1])
        if isinstance(stack[-1], str):
            word = stack.pop()
            category = stack.pop()
            tree = Tree.make_terminal(word, category)
            position += 1
        else:
            assert isinstance(stack[-1], Tree)
            children = []
            while isinstance(stack[-1], Tree):
                tree = stack.pop()
                children.append(tree)
            category = stack.pop()
            if len(children) == 1:
                tree = Tree.make_unary(category, children[0])
            elif len(children) == 2:
                right, left = children
                combinator = guess_combinator_by_triplet(
                    binary_rules, category, left.cat, right.cat
                )
                tree = Tree.make_binary(
                    category, left, right, combinator
                )
            else:
                assert False
        stack.append(tree)

    def rec() -> None:
        if len(buf) == 0:
            return
        item = buf.pop()
        assert item[0] == '(' or item[-1] == ')'
        if item[0] == '(':
            stack.append(Category.parse(item[1:]))
        elif item[-1] == ')':
            reduce(item)
        rec()

    try:
        rec()
        assert len(stack) == 1 and isinstance(stack[0], Tree)
    except AssertionError:
        raise RuntimeError('Parse failed on an invalid CCG tree')
    return stack[0], tokens


def read_ptb(filename: str) -> Iterator[ReaderResult]:
    """parse PTB-formatted file

    Args:
        filename (str): file name string

    Yields:
        Iterator[ReaderResult]: iterator object containing parse results
    """
    name0 = None
    for i, line in enumerate(open(filename)):
        line = line.strip()
        if len(line) == 0:
            continue
        if line.startswith("ID"):
            name0 = line
        else:
            tree, tokens = _parse_ptb(line)
            name = name0 or f'ID={i}'
            yield ReaderResult(name, tokens, tree)


def read_trees_guess_extension(filename: str) -> Iterator[ReaderResult]:
    """guess the file format based on the extension and parse it

    Args:
        filename (str): file name string

    Yields:
        Iterator[ReaderResult]: iterator object containing parse results
    """

    logger.info(f'reading trees from: {filename}')

    if filename.endswith('.jigg.xml'):
        logger.info('read it as jigg XML file')
        yield from read_jigg_xml(filename)

    elif filename.endswith('.xml'):
        logger.info('read it as C&C XML file')
        yield from read_xml(filename)

    elif filename.endswith('.ptb'):
        logger.info('read it as PTB format file')
        yield from read_ptb(filename)

    else:
        logger.info('read it as AUTO file')
        yield from read_auto(filename)



from typing import Dict, Any
from depccg.tree import Tree
from depccg.cat import Category


def _json_of_category(category: Category) -> Dict[str, Any]:

    def rec(node):
        if node.is_functor:
            return {
                'slash': node.slash,
                'left': rec(node.left),
                'right': rec(node.right)
            }
        else:
            feature = node.features
            return {
                'base': node.base,
                'feature': feature if len(feature) > 0 else None
            }

    return rec(category)


def json_of(
    tree: Tree,
    full: bool = False
) -> Dict[str, Any]:
    """a tree in Python dict object.

    Args:
        tree (Tree): tree object
        full (bool): whether to decomopose categories into its components, i.e., 
            {
                'slash': '/',
                'left': {'base': 'S', 'feature': 'adj'},
                'right': {'base': 'NP', 'feature': None},
            },
            or just as a string "S[adj]/NP".

    Returns:
        str: tree string in the CoNLL format
    """

    def rec(node: Tree) -> Dict[str, Any]:

        if node.is_leaf:
            res = dict(node.token)
            res['cat'] = _json_of_category(node.cat) if full else str(node.cat)
            return res
        else:
            return {
                'type': node.op_string,
                'cat': _json_of_category(node.cat) if full else str(node.cat),
                'children': [rec(child) for child in node.children]
            }

    return rec(tree)



import pytest
from depccg.cat import Category
from depccg.types import Token
from depccg.tree import Tree, ScoredTree
from depccg.printer import to_string
from depccg.instance_models import SEMANTIC_TEMPLATES
from depccg.lang import set_global_language_to


@pytest.fixture()
def en_tree():
    return Tree.make_binary(
        Category.parse("S[dcl]"),
        Tree.make_binary(
            Category.parse("S[dcl]"),
            Tree.make_binary(
                Category.parse('NP'),
                Tree.make_terminal(
                    Token(word="This", pos="DT", entity="O",
                          lemma="this", chunk="XX"),
                    Category.parse("NP[nb]/N"),
                ),
                Tree.make_terminal(
                    Token(word="paper", pos="NN", entity="O",
                          lemma="paper", chunk="XX"),
                    Category.parse("N"),
                ),
                'fa',
                '>'
            ),
            Tree.make_binary(
                Category.parse("S[dcl]\\NP"),
                Tree.make_terminal(
                    Token(word="discusses", pos="VBZ",
                          entity="O", lemma="discuss", chunk="XX"),
                    Category.parse("(S[dcl]\\NP)/NP"),
                ),
                Tree.make_binary(
                    Category.parse("NP"),
                    Tree.make_unary(
                        Category.parse("NP"),
                        Tree.make_binary(
                            Category.parse("N"),
                            Tree.make_terminal(
                                Token(word="crucial", pos="JJ", entity="O",
                                      lemma="crucial", chunk="XX"),
                                Category.parse("N/N"),
                            ),
                            Tree.make_terminal(
                                Token(word="aspects", pos="NNS", entity="O",
                                      lemma="aspect", chunk="XX"),
                                Category.parse("N"),
                            ),
                            "fa",
                            ">"
                        ),
                    ),
                    Tree.make_binary(
                        Category.parse("NP\\NP"),
                        Tree.make_terminal(
                            Token(word="of", pos="IN", entity="O",
                                  lemma="of", chunk="XX"),
                            Category.parse("(NP\\NP)/NP"),
                        ),
                        Tree.make_binary(
                            Category.parse("NP"),
                            Tree.make_terminal(
                                Token(word="this", pos="DT", entity="O",
                                      lemma="this", chunk="XX"),
                                Category.parse("NP[nb]/N"),
                            ),
                            Tree.make_binary(
                                Category.parse('N'),
                                Tree.make_terminal(
                                    Token(word="new", pos="JJ", entity="O",
                                          lemma="new", chunk="XX"),
                                    Category.parse("N/N"),
                                ),
                                Tree.make_binary(
                                    Category.parse('N'),
                                    Tree.make_terminal(
                                        Token(word="annotation", pos="NN", entity="O",
                                              lemma="annotation", chunk="XX"),
                                        Category.parse("N/N"),
                                    ),
                                    Tree.make_terminal(
                                        Token(word="scheme", pos="NN", entity="O",
                                              lemma="scheme", chunk="XX"),
                                        Category.parse("N"),
                                    ),
                                    'fa',
                                    '>'
                                ),
                                'fa',
                                '>'
                            ),
                            'fa',
                            '>'
                        ),
                        'fa',
                        '>'
                    ),
                    "ba",
                    "<"
                ),
                'fa',
                '>'
            ),
            "ba",
            "<"
        ),
        Tree.make_terminal(
            Token(word=".", pos=".", entity="O", lemma=".", chunk="XX"),
            Category.parse("."),
        ),
        "rp",
        "<rp>"
    )


@pytest.fixture()
def scored_en_tree(en_tree):
    return ScoredTree(en_tree, -0.05)


@pytest.fixture()
def ja_tree():
    return Tree.make_binary(
        Category.parse("S[mod=nm,form=base,fin=t]"),
        Tree.make_binary(
            Category.parse("S[mod=nm,form=base,fin=f]"),
            Tree.make_binary(
                Category.parse(
                    "S[mod=X1,form=X2,fin=f]/S[mod=X1,form=X2,fin=f]"),
                Tree.make_binary(
                    Category.parse("NP[case=nc,mod=nm,fin=f]"),
                    Tree.make_terminal(
                        Token(word='メロス', surf='メロス', pos='名詞', pos1='一般', pos2='*', pos3='*',
                              nflectionForm='*', inflectionType='*', reading='*', base='メロス'),
                        Category.parse("NP[case=nc,mod=nm,fin=f]")),
                    Tree.make_terminal(
                        Token(word='に', surf='に', pos='助詞', pos1='格助詞', pos2='一般', pos3='*',
                              inflectionForm='*', inflectionType='*', reading='ニ', base='に'),
                        Category.parse("NP[case=nc,mod=nm,fin=f]\\NP[case=nc,mod=nm,fin=f]")),
                    "<",
                    "<",
                ),
                Tree.make_terminal(
                    Token(word='は', surf='は', pos='助詞', pos1='係助詞', pos2='*', pos3='*',
                          nflectionForm='*', inflectionType='*', reading='ハ', base='は'),
                    Category.parse("(S[mod=X1,form=X2,fin=f]/S[mod=X1,form=X2,fin=f])\\NP[case=nc,mod=nm,fin=f]")),
                "<",
                "<",
            ),
            Tree.make_binary(
                Category.parse("S[mod=nm,form=base,fin=f]"),
                Tree.make_binary(
                    Category.parse("NP[case=ga,mod=nm,fin=f]"),
                    Tree.make_terminal(
                        Token(word='政治', surf='政治', pos='名詞', pos1='一般', pos2='*', pos3='*',
                              inflectionForm='*', inflectionType='*', reading='セイジ', base='政治'),
                        Category.parse("NP[case=nc,mod=nm,fin=f]")),
                    Tree.make_terminal(
                        Token(word='が', surf='が', pos='助詞', pos1='格助詞', pos2='一般', pos3='*',
                              nflectionForm='*', inflectionType='*', reading='ガ', base='が'),
                        Category.parse("NP[case=ga,mod=nm,fin=f]\\NP[case=nc,mod=nm,fin=f]")),
                    "<",
                    "<",
                ),
                Tree.make_binary(
                    Category.parse(
                        "S[mod=nm,form=base,fin=f]\\NP[case=ga,mod=nm,fin=f]"),
                    Tree.make_terminal(
                        Token(word='わから', surf='わから', pos='動詞', pos1='自立', pos2='*', pos3='*',
                              inflectionForm='未然形', inflectionType='五段・ラ行', reading='ワカラ', base='わかる'),
                        Category.parse("S[mod=nm,form=neg,fin=f]\\NP[case=ga,mod=nm,fin=f]")),
                    Tree.make_terminal(
                        Token(word='ぬ', surf='ぬ', pos='助動詞', pos1='*', pos2='*', pos3='*',
                              inflectionForm='基本形', inflectionType='特殊・ヌ', reading='ヌ', base='ぬ'),
                        Category.parse("S[mod=nm,form=base,fin=f]\\S[mod=nm,form=neg,fin=f]")),
                    "<B1",
                    "<B1",
                ),
                "<",
                "<",
            ),
            ">",
            ">",
        ),
        Tree.make_terminal(
            Token(word='。', surf='。', pos='記号', pos1='句点', pos2='*', pos3='*',
                  inflectionForm='*', inflectionType='*', reading='。', base='。'),
            Category.parse("S[mod=nm,form=base,fin=t]\\S[mod=nm,form=base,fin=f]")),
        "<",
        "<",
    )


@pytest.fixture()
def scored_ja_tree(ja_tree):
    return ScoredTree(ja_tree, -0.05)


def test_en_auto(scored_en_tree):

    expected = (
        'ID=1, log probability=-0.05000000\n'
        '(<T S[dcl] 0 2> (<T S[dcl] 0 2> (<T NP 0 2> (<L NP[nb]/N DT DT This NP[nb]/N>)'
        ' (<L N NN NN paper N>) ) (<T S[dcl]\\NP 0 2> (<L (S[dcl]\\NP)/NP VBZ VBZ discusses'
        ' (S[dcl]\\NP)/NP>) (<T NP 0 2> (<T NP 0 1> (<T N 0 2> (<L N/N JJ JJ crucial N/N>)'
        ' (<L N NNS NNS aspects N>) ) ) (<T NP\\NP 0 2> (<L (NP\\NP)/NP IN IN of (NP\\NP)/NP>)'
        ' (<T NP 0 2> (<L NP[nb]/N DT DT this NP[nb]/N>) (<T N 0 2> (<L N/N JJ JJ new N/N>)'
        ' (<T N 0 2> (<L N/N NN NN annotation N/N>) (<L N NN NN scheme N>) ) ) ) ) ) ) ) (<L . . . . .>) )\n'
    )

    assert to_string([scored_en_tree], format='auto') == expected


def test_en_auto_extended(scored_en_tree):
    expected = (
        'ID=1, log probability=-0.05000000\n'
        '(<T S[dcl] rp 0 2> (<T S[dcl] ba 0 2> (<T NP fa 0 2> (<L NP[nb]/N This this DT O XX NP[nb]/N>)'
        ' (<L N paper paper NN O XX N>) ) (<T S[dcl]\\NP fa 0 2> (<L (S[dcl]\\NP)/NP discusses discuss'
        ' VBZ O XX (S[dcl]\\NP)/NP>) (<T NP ba 0 2> (<T NP lex 0 1> (<T N fa 0 2> (<L N/N crucial crucial'
        ' JJ O XX N/N>) (<L N aspects aspect NNS O XX N>) ) ) (<T NP\\NP fa 0 2> (<L (NP\\NP)/NP of of IN O'
        ' XX (NP\\NP)/NP>) (<T NP fa 0 2> (<L NP[nb]/N this this DT O XX NP[nb]/N>) (<T N fa 0 2> (<L N/N'
        ' new new JJ O XX N/N>) (<T N fa 0 2> (<L N/N annotation annotation NN O XX N/N>) (<L N scheme'
        ' scheme NN O XX N>) ) ) ) ) ) ) ) (<L . . . . O XX .>) )\n'
    )
    assert to_string([scored_en_tree], format='auto_extended') == expected


def test_en_xml(scored_en_tree):
    expected = (
        '<candc>\n'
        '  <ccg sentence="1" id="1">\n'
        '    <rule type="rp" cat="S[dcl]">\n'
        '      <rule type="ba" cat="S[dcl]">\n'
        '        <rule type="fa" cat="NP">\n'
        '          <lf start="0" span="1" cat="NP[nb]/N" word="This" pos="DT" entity="O" lemma="this" chunk="XX"/>\n'
        '          <lf start="1" span="1" cat="N" word="paper" pos="NN" entity="O" lemma="paper" chunk="XX"/>\n'
        '        </rule>\n'
        '        <rule type="fa" cat="S[dcl]\\NP">\n'
        '          <lf start="2" span="1" cat="(S[dcl]\\NP)/NP" word="discusses" pos="VBZ" entity="O" lemma="discuss" chunk="XX"/>\n'
        '          <rule type="ba" cat="NP">\n'
        '            <rule type="lex" cat="NP">\n'
        '              <rule type="fa" cat="N">\n'
        '                <lf start="3" span="1" cat="N/N" word="crucial" pos="JJ" entity="O" lemma="crucial" chunk="XX"/>\n'
        '                <lf start="4" span="1" cat="N" word="aspects" pos="NNS" entity="O" lemma="aspect" chunk="XX"/>\n'
        '              </rule>\n'
        '            </rule>\n'
        '            <rule type="fa" cat="NP\\NP">\n'
        '              <lf start="5" span="1" cat="(NP\\NP)/NP" word="of" pos="IN" entity="O" lemma="of" chunk="XX"/>\n'
        '              <rule type="fa" cat="NP">\n'
        '                <lf start="6" span="1" cat="NP[nb]/N" word="this" pos="DT" entity="O" lemma="this" chunk="XX"/>\n'
        '                <rule type="fa" cat="N">\n'
        '                  <lf start="7" span="1" cat="N/N" word="new" pos="JJ" entity="O" lemma="new" chunk="XX"/>\n'
        '                  <rule type="fa" cat="N">\n'
        '                    <lf start="8" span="1" cat="N/N" word="annotation" pos="NN" entity="O" lemma="annotation" chunk="XX"/>\n'
        '                    <lf start="9" span="1" cat="N" word="scheme" pos="NN" entity="O" lemma="scheme" chunk="XX"/>\n'
        '                  </rule>\n'
        '                </rule>\n'
        '              </rule>\n'
        '            </rule>\n'
        '          </rule>\n'
        '        </rule>\n'
        '      </rule>\n'
        '      <lf start="10" span="1" cat="." word="." pos="." entity="O" lemma="." chunk="XX"/>\n'
        '    </rule>\n'
        '  </ccg>\n'
        '</candc>\n'
    )
    assert to_string([scored_en_tree], format='xml') == expected


def test_en_deriv(scored_en_tree):

    expected = (
        'ID=1, log probability=-0.05000000\n'
        ' NP[nb]/N    N    (S[dcl]\\NP)/NP    N/N       N     (NP\\NP)/NP  NP[nb]/N  N/N     N/N        N     .\n'
        '   This    paper    discusses     crucial  aspects      of        this    new  annotation  scheme  .\n'
        '----------------->\n'
        '       NP\n'
        '                                 ------------------>\n'
        '                                         N\n'
        '                                 ------------------<un>\n'
        '                                         NP\n'
        '                                                                              -------------------->\n'
        '                                                                                       N\n'
        '                                                                         ------------------------->\n'
        '                                                                                     N\n'
        '                                                               ----------------------------------->\n'
        '                                                                               NP\n'
        '                                                   ----------------------------------------------->\n'
        '                                                                        NP\\NP\n'
        '                                 -----------------------------------------------------------------<\n'
        '                                                                NP\n'
        '                 --------------------------------------------------------------------------------->\n'
        '                                                     S[dcl]\\NP\n'
        '--------------------------------------------------------------------------------------------------<\n'
        '                                              S[dcl]\n'
        '-----------------------------------------------------------------------------------------------------<rp>\n'
        '                                               S[dcl]\n\n'
    )

    assert to_string([scored_en_tree], format='deriv') == expected


def test_en_prolog(scored_en_tree):

    expected = (
        ":- op(601, xfx, (/)).\n"
        ":- op(601, xfx, (\\)).\n"
        ":- multifile ccg/2, id/2.\n"
        ":- discontiguous ccg/2, id/2.\n"
        "\n"
        "ccg(1,\n"
        " rp(s:dcl,\n"
        "  ba(s:dcl,\n"
        "   fa(np,\n"
        "    t((np:nb/n), 'This', 'this', 'DT', 'XX', 'O'),\n"
        "    t(n, 'paper', 'paper', 'NN', 'XX', 'O')),\n"
        "   fa((s:dcl\\np),\n"
        "    t(((s:dcl\\np)/np), 'discusses', 'discuss', 'VBZ', 'XX', 'O'),\n"
        "    ba(np,\n"
        "     lx(np, n,\n"
        "      fa(n,\n"
        "       t((n/n), 'crucial', 'crucial', 'JJ', 'XX', 'O'),\n"
        "       t(n, 'aspects', 'aspect', 'NNS', 'XX', 'O'))),\n"
        "     fa((np\\np),\n"
        "      t(((np\\np)/np), 'of', 'of', 'IN', 'XX', 'O'),\n"
        "      fa(np,\n"
        "       t((np:nb/n), 'this', 'this', 'DT', 'XX', 'O'),\n"
        "       fa(n,\n"
        "        t((n/n), 'new', 'new', 'JJ', 'XX', 'O'),\n"
        "        fa(n,\n"
        "         t((n/n), 'annotation', 'annotation', 'NN', 'XX', 'O'),\n"
        "         t(n, 'scheme', 'scheme', 'NN', 'XX', 'O')))))))),\n"
        "  t(period, '.', '.', '.', 'XX', 'O'))).\n\n"
    )

    assert to_string([scored_en_tree], format='prolog') == expected


def test_en_jigg_xml(scored_en_tree):

    expected = (
        '<root>\n'
        '  <document>\n'
        '    <sentences>\n'
        '      <sentence>\n'
        '        <tokens>\n'
        '          <token start="0" cat="NP[nb]/N" id="s0_0" pos="DT" entity="O" chunk="XX" surf="This" base="this"/>\n'
        '          <token start="1" cat="N" id="s0_1" pos="NN" entity="O" chunk="XX" surf="paper" base="paper"/>\n'
        '          <token start="2" cat="(S[dcl]\\NP)/NP" id="s0_2" pos="VBZ" entity="O" chunk="XX" surf="discusses" base="discuss"/>\n'
        '          <token start="3" cat="N/N" id="s0_3" pos="JJ" entity="O" chunk="XX" surf="crucial" base="crucial"/>\n'
        '          <token start="4" cat="N" id="s0_4" pos="NNS" entity="O" chunk="XX" surf="aspects" base="aspect"/>\n'
        '          <token start="5" cat="(NP\\NP)/NP" id="s0_5" pos="IN" entity="O" chunk="XX" surf="of" base="of"/>\n'
        '          <token start="6" cat="NP[nb]/N" id="s0_6" pos="DT" entity="O" chunk="XX" surf="this" base="this"/>\n'
        '          <token start="7" cat="N/N" id="s0_7" pos="JJ" entity="O" chunk="XX" surf="new" base="new"/>\n'
        '          <token start="8" cat="N/N" id="s0_8" pos="NN" entity="O" chunk="XX" surf="annotation" base="annotation"/>\n'
        '          <token start="9" cat="N" id="s0_9" pos="NN" entity="O" chunk="XX" surf="scheme" base="scheme"/>\n'
        '          <token start="10" cat="." id="s0_10" pos="." entity="O" chunk="XX" surf="." base="."/>\n'
        '        </tokens>\n'
        '        <ccg id="s0_ccg0" root="s0_sp0" score="-0.05">\n'
        '          <span category="S[dcl=true]" id="s0_sp0" child="s0_sp1 s0_sp21" rule="rp" begin="0" end="11" root="true"/>\n'
        '          <span category="S[dcl=true]" id="s0_sp1" child="s0_sp2 s0_sp5" rule="ba" begin="0" end="10"/>\n'
        '          <span category="NP" id="s0_sp2" child="s0_sp3 s0_sp4" rule="fa" begin="0" end="2"/>\n'
        '          <span category="NP[nb=true]/N" id="s0_sp3" terminal="s0_0" begin="0" end="1"/>\n'
        '          <span category="N" id="s0_sp4" terminal="s0_1" begin="1" end="2"/>\n'
        '          <span category="S[dcl=true]\\NP" id="s0_sp5" child="s0_sp6 s0_sp7" rule="fa" begin="2" end="10"/>\n'
        '          <span category="(S[dcl=true]\\NP)/NP" id="s0_sp6" terminal="s0_2" begin="2" end="3"/>\n'
        '          <span category="NP" id="s0_sp7" child="s0_sp8 s0_sp12" rule="ba" begin="3" end="10"/>\n'
        '          <span category="NP" id="s0_sp8" child="s0_sp9" rule="lex" begin="3" end="5"/>\n'
        '          <span category="N" id="s0_sp9" child="s0_sp10 s0_sp11" rule="fa" begin="3" end="5"/>\n'
        '          <span category="N/N" id="s0_sp10" terminal="s0_3" begin="3" end="4"/>\n'
        '          <span category="N" id="s0_sp11" terminal="s0_4" begin="4" end="5"/>\n'
        '          <span category="NP\\NP" id="s0_sp12" child="s0_sp13 s0_sp14" rule="fa" begin="5" end="10"/>\n'
        '          <span category="(NP\\NP)/NP" id="s0_sp13" terminal="s0_5" begin="5" end="6"/>\n'
        '          <span category="NP" id="s0_sp14" child="s0_sp15 s0_sp16" rule="fa" begin="6" end="10"/>\n'
        '          <span category="NP[nb=true]/N" id="s0_sp15" terminal="s0_6" begin="6" end="7"/>\n'
        '          <span category="N" id="s0_sp16" child="s0_sp17 s0_sp18" rule="fa" begin="7" end="10"/>\n'
        '          <span category="N/N" id="s0_sp17" terminal="s0_7" begin="7" end="8"/>\n'
        '          <span category="N" id="s0_sp18" child="s0_sp19 s0_sp20" rule="fa" begin="8" end="10"/>\n'
        '          <span category="N/N" id="s0_sp19" terminal="s0_8" begin="8" end="9"/>\n'
        '          <span category="N" id="s0_sp20" terminal="s0_9" begin="9" end="10"/>\n'
        '          <span category="." id="s0_sp21" terminal="s0_10" begin="10" end="11"/>\n'
        '        </ccg>\n'
        '      </sentence>\n'
        '    </sentences>\n'
        '  </document>\n'
        '</root>\n'
    )

    assert to_string([scored_en_tree], format='jigg_xml') == expected


def test_en_jigg_xml_ccg2lambda(scored_en_tree):

    expected = (
        '<?xml version=\'1.0\' encoding=\'utf-8\'?>\n'
        '<root>\n'
        '  <document>\n'
        '    <sentences>\n'
        '      <sentence>\n'
        '        <tokens>\n'
        '          <token start="0" cat="NP[nb]/N" id="s0_0" pos="DT" entity="O" chunk="XX" surf="This" base="this"/>\n'
        '          <token start="1" cat="N" id="s0_1" pos="NN" entity="O" chunk="XX" surf="paper" base="paper"/>\n'
        '          <token start="2" cat="(S[dcl]\\NP)/NP" id="s0_2" pos="VBZ" entity="O" chunk="XX" surf="discusses" base="discuss"/>\n'
        '          <token start="3" cat="N/N" id="s0_3" pos="JJ" entity="O" chunk="XX" surf="crucial" base="crucial"/>\n'
        '          <token start="4" cat="N" id="s0_4" pos="NNS" entity="O" chunk="XX" surf="aspects" base="aspect"/>\n'
        '          <token start="5" cat="(NP\\NP)/NP" id="s0_5" pos="IN" entity="O" chunk="XX" surf="of" base="of"/>\n'
        '          <token start="6" cat="NP[nb]/N" id="s0_6" pos="DT" entity="O" chunk="XX" surf="this" base="this"/>\n'
        '          <token start="7" cat="N/N" id="s0_7" pos="JJ" entity="O" chunk="XX" surf="new" base="new"/>\n'
        '          <token start="8" cat="N/N" id="s0_8" pos="NN" entity="O" chunk="XX" surf="annotation" base="annotation"/>\n'
        '          <token start="9" cat="N" id="s0_9" pos="NN" entity="O" chunk="XX" surf="scheme" base="scheme"/>\n'
        '          <token start="10" cat="." id="s0_10" pos="." entity="O" chunk="XX" surf="." base="."/>\n'
        '        </tokens>\n'
        '        <ccg id="s0_ccg0" root="s0_sp0" score="-0.05">\n'
        '          <span category="S[dcl=true]" id="s0_sp0" child="s0_sp1 s0_sp21" rule="rp" begin="0" end="11" root="true"/>\n'
        '          <span category="S[dcl=true]" id="s0_sp1" child="s0_sp2 s0_sp5" rule="ba" begin="0" end="10"/>\n'
        '          <span category="NP" id="s0_sp2" child="s0_sp3 s0_sp4" rule="fa" begin="0" end="2"/>\n'
        '          <span category="NP[nb=true]/N" id="s0_sp3" terminal="s0_0" begin="0" end="1"/>\n'
        '          <span category="N" id="s0_sp4" terminal="s0_1" begin="1" end="2"/>\n'
        '          <span category="S[dcl=true]\\NP" id="s0_sp5" child="s0_sp6 s0_sp7" rule="fa" begin="2" end="10"/>\n'
        '          <span category="(S[dcl=true]\\NP)/NP" id="s0_sp6" terminal="s0_2" begin="2" end="3"/>\n'
        '          <span category="NP" id="s0_sp7" child="s0_sp8 s0_sp12" rule="ba" begin="3" end="10"/>\n'
        '          <span category="NP" id="s0_sp8" child="s0_sp9" rule="lex" begin="3" end="5"/>\n'
        '          <span category="N" id="s0_sp9" child="s0_sp10 s0_sp11" rule="fa" begin="3" end="5"/>\n'
        '          <span category="N/N" id="s0_sp10" terminal="s0_3" begin="3" end="4"/>\n'
        '          <span category="N" id="s0_sp11" terminal="s0_4" begin="4" end="5"/>\n'
        '          <span category="NP\\NP" id="s0_sp12" child="s0_sp13 s0_sp14" rule="fa" begin="5" end="10"/>\n'
        '          <span category="(NP\\NP)/NP" id="s0_sp13" terminal="s0_5" begin="5" end="6"/>\n'
        '          <span category="NP" id="s0_sp14" child="s0_sp15 s0_sp16" rule="fa" begin="6" end="10"/>\n'
        '          <span category="NP[nb=true]/N" id="s0_sp15" terminal="s0_6" begin="6" end="7"/>\n'
        '          <span category="N" id="s0_sp16" child="s0_sp17 s0_sp18" rule="fa" begin="7" end="10"/>\n'
        '          <span category="N/N" id="s0_sp17" terminal="s0_7" begin="7" end="8"/>\n'
        '          <span category="N" id="s0_sp18" child="s0_sp19 s0_sp20" rule="fa" begin="8" end="10"/>\n'
        '          <span category="N/N" id="s0_sp19" terminal="s0_8" begin="8" end="9"/>\n'
        '          <span category="N" id="s0_sp20" terminal="s0_9" begin="9" end="10"/>\n'
        '          <span category="." id="s0_sp21" terminal="s0_10" begin="10" end="11"/>\n'
        '        </ccg>\n'
        '        <semantics status="success" ccg_id="s0_ccg0" root="s0_sp0">\n'
        '          <span id="s0_sp0" child="s0_sp1 s0_sp21" sem="exists x.(_paper(x) &amp; True &amp; exists z2.(_aspect(z2) &amp; _crucial(z2) &amp; exists z1.(_scheme(z1) &amp; _annotation(z1) &amp; _new(z1) &amp; True &amp; (z2 = z1)) &amp; True &amp; exists e.(_discuss(e) &amp; (Subj(e) = x) &amp; (Acc(e) = z2) &amp; True)))"/>\n'
        '          <span id="s0_sp1" child="s0_sp2 s0_sp5" sem="\\K.exists x.(_paper(x) &amp; True &amp; exists z2.(_aspect(z2) &amp; _crucial(z2) &amp; exists z1.(_scheme(z1) &amp; _annotation(z1) &amp; _new(z1) &amp; True &amp; (z2 = z1)) &amp; True &amp; exists e.(_discuss(e) &amp; (Subj(e) = x) &amp; (Acc(e) = z2) &amp; K(e))))"/>\n'
        '          <span id="s0_sp2" child="s0_sp3 s0_sp4" sem="\\F2 F3.exists x.(_paper(x) &amp; F2(x) &amp; F3(x))"/>\n'
        '          <span id="s0_sp3" sem="\\F1 F2 F3.exists x.(F1(x) &amp; F2(x) &amp; F3(x))"/>\n'
        '          <span id="s0_sp4" sem="\\x._paper(x)" type="_paper : Entity -&gt; Prop"/>\n'
        '          <span id="s0_sp5" child="s0_sp6 s0_sp7" sem="\\Q2 K.Q2(\\x.True,\\x.exists z2.(_aspect(z2) &amp; _crucial(z2) &amp; exists z1.(_scheme(z1) &amp; _annotation(z1) &amp; _new(z1) &amp; True &amp; (z2 = z1)) &amp; True &amp; exists e.(_discuss(e) &amp; (Subj(e) = x) &amp; (Acc(e) = z2) &amp; K(e))))"/>\n'
        '          <span id="s0_sp6" sem="\\Q1 Q2 K.Q2(\\x.True,\\x.Q1(\\y.True,\\y.exists e.(_discuss(e) &amp; (Subj(e) = x) &amp; (Acc(e) = y) &amp; K(e))))" type="_discuss : Event -&gt; Prop"/>\n'
        '          <span id="s0_sp7" child="s0_sp8 s0_sp12" sem="\\F1 F2.exists x.(_aspect(x) &amp; _crucial(x) &amp; exists z1.(_scheme(z1) &amp; _annotation(z1) &amp; _new(z1) &amp; True &amp; (x = z1)) &amp; F1(x) &amp; F2(x))"/>\n'
        '          <span id="s0_sp8" child="s0_sp9" sem="\\F1 F2.exists x.(_aspect(x) &amp; _crucial(x) &amp; F1(x) &amp; F2(x))"/>\n'
        '          <span id="s0_sp9" child="s0_sp10 s0_sp11" sem="\\x.(_aspect(x) &amp; _crucial(x))"/>\n'
        '          <span id="s0_sp10" sem="\\F x.(F(x) &amp; _crucial(x))" type="_crucial : Entity -&gt; Prop"/>\n'
        '          <span id="s0_sp11" sem="\\x._aspect(x)" type="_aspect : Entity -&gt; Prop"/>\n'
        '          <span id="s0_sp12" child="s0_sp13 s0_sp14" sem="\\Q2 F1 F2.Q2(\\x.(exists z1.(_scheme(z1) &amp; _annotation(z1) &amp; _new(z1) &amp; True &amp; (x = z1)) &amp; F1(x)),F2)"/>\n'
        '          <span id="s0_sp13" sem="\\Q1 Q2 F1 F2.Q2(\\x.(Q1(\\w.True,\\y.(x = y)) &amp; F1(x)),F2)"/>\n'
        '          <span id="s0_sp14" child="s0_sp15 s0_sp16" sem="\\F2 F3.exists x.(_scheme(x) &amp; _annotation(x) &amp; _new(x) &amp; F2(x) &amp; F3(x))"/>\n'
        '          <span id="s0_sp15" sem="\\F1 F2 F3.exists x.(F1(x) &amp; F2(x) &amp; F3(x))"/>\n'
        '          <span id="s0_sp16" child="s0_sp17 s0_sp18" sem="\\x.(_scheme(x) &amp; _annotation(x) &amp; _new(x))"/>\n'
        '          <span id="s0_sp17" sem="\\F x.(F(x) &amp; _new(x))" type="_new : Entity -&gt; Prop"/>\n'
        '          <span id="s0_sp18" child="s0_sp19 s0_sp20" sem="\\x.(_scheme(x) &amp; _annotation(x))"/>\n'
        '          <span id="s0_sp19" sem="\\F x.(F(x) &amp; _annotation(x))" type="_annotation : Entity -&gt; Prop"/>\n'
        '          <span id="s0_sp20" sem="\\x._scheme(x)" type="_scheme : Entity -&gt; Prop"/>\n'
        '          <span id="s0_sp21" sem="\\X.X"/>\n'
        '        </semantics>\n'
        '      </sentence>\n'
        '    </sentences>\n'
        '  </document>\n'
        '</root>\n'
    )

    assert (
        to_string(
            [scored_en_tree],
            format='jigg_xml_ccg2lambda',
            semantic_templates=SEMANTIC_TEMPLATES['en'],
        ) == expected
    )


def test_en_conll(scored_en_tree):

    expected = (
        '# ID=1\n'
        '# log probability=-0.05000000\n'
        '1	This	this	DT	DT	_	0	NP[nb]/N	_	(<T S[dcl] 0 2> (<T S[dcl] 0 2> (<T NP 0 2> (<L NP[nb]/N DT DT This NP[nb]/N>)\n'
        '2	paper	paper	NN	NN	_	1	N	_	(<L N NN NN paper N>) )\n'
        '3	discusses	discuss	VBZ	VBZ	_	1	(S[dcl]\\NP)/NP	_	(<T S[dcl]\\NP 0 2> (<L (S[dcl]\\NP)/NP VBZ VBZ discusses (S[dcl]\\NP)/NP>)\n'
        '4	crucial	crucial	JJ	JJ	_	3	N/N	_	(<T NP 0 2> (<T NP 0 1> (<T N 0 2> (<L N/N JJ JJ crucial N/N>)\n'
        '5	aspects	aspect	NNS	NNS	_	4	N	_	(<L N NNS NNS aspects N>) ) )\n'
        '6	of	of	IN	IN	_	4	(NP\\NP)/NP	_	(<T NP\\NP 0 2> (<L (NP\\NP)/NP IN IN of (NP\\NP)/NP>)\n'
        '7	this	this	DT	DT	_	6	NP[nb]/N	_	(<T NP 0 2> (<L NP[nb]/N DT DT this NP[nb]/N>)\n'
        '8	new	new	JJ	JJ	_	7	N/N	_	(<T N 0 2> (<L N/N JJ JJ new N/N>)\n'
        '9	annotation	annotation	NN	NN	_	8	N/N	_	(<T N 0 2> (<L N/N NN NN annotation N/N>)\n'
        '10	scheme	scheme	NN	NN	_	9	N	_	(<L N NN NN scheme N>) ) ) ) ) ) ) )\n'
        '11	.	.	.	.	_	1	.	_	(<L . . . . .>) )\n'
    )

    assert to_string([scored_en_tree], format='conll') == expected


def test_en_json(scored_en_tree):

    expected = (
        '{\n'
        '    "1": [\n'
        '        {\n'
        '            "type": "rp",\n'
        '            "cat": "S[dcl]",\n'
        '            "children": [\n'
        '                {\n'
        '                    "type": "ba",\n'
        '                    "cat": "S[dcl]",\n'
        '                    "children": [\n'
        '                        {\n'
        '                            "type": "fa",\n'
        '                            "cat": "NP",\n'
        '                            "children": [\n'
        '                                {\n'
        '                                    "word": "This",\n'
        '                                    "pos": "DT",\n'
        '                                    "entity": "O",\n'
        '                                    "lemma": "this",\n'
        '                                    "chunk": "XX",\n'
        '                                    "cat": "NP[nb]/N"\n'
        '                                },\n'
        '                                {\n'
        '                                    "word": "paper",\n'
        '                                    "pos": "NN",\n'
        '                                    "entity": "O",\n'
        '                                    "lemma": "paper",\n'
        '                                    "chunk": "XX",\n'
        '                                    "cat": "N"\n'
        '                                }\n'
        '                            ]\n'
        '                        },\n'
        '                        {\n'
        '                            "type": "fa",\n'
        '                            "cat": "S[dcl]\\\\NP",\n'
        '                            "children": [\n'
        '                                {\n'
        '                                    "word": "discusses",\n'
        '                                    "pos": "VBZ",\n'
        '                                    "entity": "O",\n'
        '                                    "lemma": "discuss",\n'
        '                                    "chunk": "XX",\n'
        '                                    "cat": "(S[dcl]\\\\NP)/NP"\n'
        '                                },\n'
        '                                {\n'
        '                                    "type": "ba",\n'
        '                                    "cat": "NP",\n'
        '                                    "children": [\n'
        '                                        {\n'
        '                                            "type": "lex",\n'
        '                                            "cat": "NP",\n'
        '                                            "children": [\n'
        '                                                {\n'
        '                                                    "type": "fa",\n'
        '                                                    "cat": "N",\n'
        '                                                    "children": [\n'
        '                                                        {\n'
        '                                                            "word": "crucial",\n'
        '                                                            "pos": "JJ",\n'
        '                                                            "entity": "O",\n'
        '                                                            "lemma": "crucial",\n'
        '                                                            "chunk": "XX",\n'
        '                                                            "cat": "N/N"\n'
        '                                                        },\n'
        '                                                        {\n'
        '                                                            "word": "aspects",\n'
        '                                                            "pos": "NNS",\n'
        '                                                            "entity": "O",\n'
        '                                                            "lemma": "aspect",\n'
        '                                                            "chunk": "XX",\n'
        '                                                            "cat": "N"\n'
        '                                                        }\n'
        '                                                    ]\n'
        '                                                }\n'
        '                                            ]\n'
        '                                        },\n'
        '                                        {\n'
        '                                            "type": "fa",\n'
        '                                            "cat": "NP\\\\NP",\n'
        '                                            "children": [\n'
        '                                                {\n'
        '                                                    "word": "of",\n'
        '                                                    "pos": "IN",\n'
        '                                                    "entity": "O",\n'
        '                                                    "lemma": "of",\n'
        '                                                    "chunk": "XX",\n'
        '                                                    "cat": "(NP\\\\NP)/NP"\n'
        '                                                },\n'
        '                                                {\n'
        '                                                    "type": "fa",\n'
        '                                                    "cat": "NP",\n'
        '                                                    "children": [\n'
        '                                                        {\n'
        '                                                            "word": "this",\n'
        '                                                            "pos": "DT",\n'
        '                                                            "entity": "O",\n'
        '                                                            "lemma": "this",\n'
        '                                                            "chunk": "XX",\n'
        '                                                            "cat": "NP[nb]/N"\n'
        '                                                        },\n'
        '                                                        {\n'
        '                                                            "type": "fa",\n'
        '                                                            "cat": "N",\n'
        '                                                            "children": [\n'
        '                                                                {\n'
        '                                                                    "word": "new",\n'
        '                                                                    "pos": "JJ",\n'
        '                                                                    "entity": "O",\n'
        '                                                                    "lemma": "new",\n'
        '                                                                    "chunk": "XX",\n'
        '                                                                    "cat": "N/N"\n'
        '                                                                },\n'
        '                                                                {\n'
        '                                                                    "type": "fa",\n'
        '                                                                    "cat": "N",\n'
        '                                                                    "children": [\n'
        '                                                                        {\n'
        '                                                                            "word": "annotation",\n'
        '                                                                            "pos": "NN",\n'
        '                                                                            "entity": "O",\n'
        '                                                                            "lemma": "annotation",\n'
        '                                                                            "chunk": "XX",\n'
        '                                                                            "cat": "N/N"\n'
        '                                                                        },\n'
        '                                                                        {\n'
        '                                                                            "word": "scheme",\n'
        '                                                                            "pos": "NN",\n'
        '                                                                            "entity": "O",\n'
        '                                                                            "lemma": "scheme",\n'
        '                                                                            "chunk": "XX",\n'
        '                                                                            "cat": "N"\n'
        '                                                                        }\n'
        '                                                                    ]\n'
        '                                                                }\n'
        '                                                            ]\n'
        '                                                        }\n'
        '                                                    ]\n'
        '                                                }\n'
        '                                            ]\n'
        '                                        }\n'
        '                                    ]\n'
        '                                }\n'
        '                            ]\n'
        '                        }\n'
        '                    ]\n'
        '                },\n'
        '                {\n'
        '                    "word": ".",\n'
        '                    "pos": ".",\n'
        '                    "entity": "O",\n'
        '                    "lemma": ".",\n'
        '                    "chunk": "XX",\n'
        '                    "cat": "."\n'
        '                }\n'
        '            ],\n'
        '            "log_prob": -0.05\n'
        '        }\n'
        '    ]\n'
        '}'
    )

    assert to_string([scored_en_tree], format='json') == expected


def test_en_ptb(scored_en_tree):

    expected = (
        'ID=1, log probability=-0.05000000\n'
        '(ROOT (S[dcl] (S[dcl] (NP (NP[nb]/N This) (N paper)) (S[dcl]\\NP ((S[dcl]\\NP)/NP discusses)'
        ' (NP (NP (N (N/N crucial) (N aspects))) (NP\\NP ((NP\\NP)/NP of) (NP (NP[nb]/N this) (N (N/N new)'
        ' (N (N/N annotation) (N scheme)))))))) (. .)))\n'
    )

    assert to_string([scored_en_tree], format='ptb') == expected


def test_ja_ja(scored_ja_tree):

    expected = (
        'ID=1, log probability=-0.05000000\n'
        '{< S[mod=nm,form=base,fin=t] {> S[mod=nm,form=base,fin=f] {< S[mod=X1,form=X2,fin=f]/S[mod=X1,form=X2,fin=f]'
        ' {< NP[case=nc,mod=nm,fin=f] {NP[case=nc,mod=nm,fin=f] メロス/メロス/名詞-一般/_}'
        ' {NP[case=nc,mod=nm,fin=f]\\NP[case=nc,mod=nm,fin=f] に/に/助詞-格助詞-一般/_}}'
        ' {(S[mod=X1,form=X2,fin=f]/S[mod=X1,form=X2,fin=f])\\NP[case=nc,mod=nm,fin=f] は/は/助詞-係助詞/_}}'
        ' {< S[mod=nm,form=base,fin=f] {< NP[case=ga,mod=nm,fin=f] {NP[case=nc,mod=nm,fin=f] 政治/政治/名詞-一般/_}'
        ' {NP[case=ga,mod=nm,fin=f]\\NP[case=nc,mod=nm,fin=f] が/が/助詞-格助詞-一般/_}}'
        ' {<B1 S[mod=nm,form=base,fin=f]\\NP[case=ga,mod=nm,fin=f] {S[mod=nm,form=neg,fin=f]\\NP[case=ga,mod=nm,fin=f]'
        ' わから/わから/動詞-自立/未然形-五段・ラ行} {S[mod=nm,form=base,fin=f]\\S[mod=nm,form=neg,fin=f]'
        ' ぬ/ぬ/助動詞/基本形-特殊・ヌ}}}} {S[mod=nm,form=base,fin=t]\\S[mod=nm,form=base,fin=f] 。/。/記号-句点/_}}\n'
    )

    assert to_string([scored_ja_tree], format='ja') == expected


def test_ja_prolog(scored_ja_tree):

    expected = (
        ":- op(601, xfx, (/)).\n"
        ":- op(601, xfx, (\\)).\n"
        ":- multifile ccg/2, id/2.\n"
        ":- discontiguous ccg/2, id/2.\n"
        "\n"
        "ccg(1,\n"
        " ba(s,\n"
        "  fa(s,\n"
        "   ba((s/s),\n"
        "    ba(np:nc,\n"
        "     t(np:nc, 'メロス', 'メロス', '名詞/一般/*/*', '*', '*'),\n"
        "     t((np:nc\\np:nc), 'に', 'に', '助詞/格助詞/一般/*', '*', '*')),\n"
        "    t(((s/s)\\np:nc), 'は', 'は', '助詞/係助詞/*/*', '*', '*')),\n"
        "   ba(s,\n"
        "    ba(np:ga,\n"
        "     t(np:nc, '政治', '政治', '名詞/一般/*/*', '*', '*'),\n"
        "     t((np:ga\\np:nc), 'が', 'が', '助詞/格助詞/一般/*', '*', '*')),\n"
        "    bc1((s\\np:ga),\n"
        "     t((s\\np:ga), 'わから', 'わかる', '動詞/自立/*/*', '未然形', '五段・ラ行'),\n"
        "     t((s\\s), 'ぬ', 'ぬ', '助動詞/*/*/*', '基本形', '特殊・ヌ')))),\n"
        "  t((s\\s), '。', '。', '記号/句点/*/*', '*', '*'))).\n\n"
    )
    set_global_language_to('ja')
    assert to_string([scored_ja_tree], format='prolog') == expected



import argparse

from depccg.tools.reader import read_trees_guess_extension
from depccg.printer import print_
from depccg.lang import set_global_language_to
from depccg.annotator import english_annotator
from depccg.instance_models import SEMANTIC_TEMPLATES
from depccg.types import ScoringResult

LANG = 'en'

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'PATH',
        help='path to either of *.auto, *.xml, *.jigg.xml, *.ptb')
    parser.add_argument(
        '--annotator',
        default='spacy',
        choices=english_annotator.keys(),
        help='annotate POS, named entity, and lemmas using this library')
    parser.add_argument(
        '-f',
        '--format',
        default='xml',
        choices=[
            'auto', 'xml', 'prolog', 'jigg_xml',
            'jigg_xml_ccg2lambda', 'json'],
        help='output format')
    parser.add_argument(
        '--semantic-templates',
        help='semantic templates used in "ccg2lambda" format output')
    args = parser.parse_args()

    trees = [
        [ScoringResult(tree, 0.0)]
        for _, _, tree in read_trees_guess_extension(args.PATH)
    ]

    set_global_language_to(LANG)
    semantic_templates = args.semantic_templates or SEMANTIC_TEMPLATES[LANG]
    print_(
        trees,
        format=args.format,
        semantic_templates=semantic_templates
    )



# -*- coding: utf-8 -*-
#
#  Copyright 2015 Pascual Martinez-Gomez
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import re

from nltk.sem.logic import Expression

from .category import Category
from .logic_parser import lexpr
from .normalization import normalize_token

class SemanticRule(object):
    def __init__(self, category, semantics, attributes = {}):
        if not isinstance(category, Category):
            self.category = Category(category)
        else:
            self.category = category
        if semantics and not isinstance(semantics, Expression):
            self.semantics = lexpr(semantics)
        else:
            self.semantics = semantics
        self.attributes = copy.deepcopy(attributes)
        if 'surf' in self.attributes:
          self.attributes['surf'] = normalize_token(self.attributes['surf'])
        if 'base' in self.attributes:
          self.attributes['base'] = normalize_token(self.attributes['base'])

    def match(self, other):
        # Check class membership and special attribute matches.
        if not isinstance(other, self.__class__) \
           or not isinstance(other.category, self.category.__class__) \
           or not self.category.match(other.category):
            return False
        # If one rule is terminal but not the other, then they do not match.
        if self.is_terminal_rule() != other.is_terminal_rule():
            return False
        # Obtain generic list of attributes from both semantic rules,
        # and check whether they match each other (or are underspecified).
        self_attr_names = set(self.attributes.keys())
        other_attr_names = set(other.attributes.keys())
        attribute_names = self_attr_names.union(other_attr_names)
        attribute_names = self.remove_control_attribute_names(attribute_names)
        for attribute_name in attribute_names:
            self_attr_value = self.attributes.get(attribute_name)
            other_attr_value = other.attributes.get(attribute_name)
            if not attributes_match(attribute_name, self_attr_value, other_attr_value):
                return False
        if not wildcard_match(attribute_names, self.attributes, other.attributes):
            return False
        return True

    def remove_control_attribute_names(self, attribute_names):
        control_attrs = ['var_paths']
        return [a for a in attribute_names if a not in control_attrs]

    def is_terminal_rule(self):
        if 'rule' in self.attributes:
            return False
        for attribute_name in self.attributes:
            if attribute_name.startswith('child'):
                return False
        return True

def attributes_match(attribute_name, src_attr_value, trg_attr_value):
    # If this attribute is an arbitrary type specification, it doesn't count
    # as a rule attribute to match against the CCG tree. Thus, return True.
    if 'coq_type' in attribute_name:
        return True
    # If the attribute is a wildcard, we do not check for match here. return True.
    if '_any_' in attribute_name:
        return True
    if src_attr_value is not None and trg_attr_value is None:
        return False
    if src_attr_value is None and trg_attr_value is not None:
        return True
    if src_attr_value is None and trg_attr_value is None:
        return True
    # Case: src_attr_value is not None and trg_attr_value is not None
    if not 'category' in attribute_name:
        return src_attr_value.lower() == trg_attr_value.lower()
    # Comparing categories needs feature unification:
    src_category = Category(src_attr_value)
    trg_category = Category(trg_attr_value)
    return src_category.match(trg_category)

def any_attribute_matches(attribute_name, src_attributes, trg_attributes):
    wildcard_names = re.findall(r'_any_(.*)', attribute_name)
    assert wildcard_names, 'Attribute name invalid: {0}'.format(attribute_name)
    wildcard_name = wildcard_names[0]
    src_attr_value = src_attributes.get(attribute_name, None)
    assert src_attr_value
    trg_attr_values = [value for key, value in trg_attributes.items() \
                       if key.endswith(wildcard_name)]
    for trg_attr_value in trg_attr_values:
        if wildcard_name == 'category':
            src_category = Category(src_attr_value)
            trg_category = Category(trg_attr_value)
            if src_category.match(trg_category):
                return True
        else:
            if src_attr_value.lower() == trg_attr_value.lower():
                return True
    return False

def wildcard_match(attribute_names, src_attributes, trg_attributes):
    wildcard_attributes = []
    # Obtain attributes specified as a wildcard "any"
    for attribute_name in src_attributes.keys():
        if re.findall(r'_any_', attribute_name):
            wildcard_attributes.append(attribute_name)
    # match of attributes specified by wildcards.
    for attribute_name in wildcard_attributes:
        if not any_attribute_matches(attribute_name, src_attributes, trg_attributes):
            return False
    return True



from depccg.lang import get_global_language
from typing import NamedTuple, List, Iterator, Union
from depccg.cat import Category
from depccg.grammar import guess_combinator_by_triplet, en, ja
from depccg.types import Token


BINARY_RULES = {
    'en': en.apply_binary_rules,
    'ja': ja.apply_binary_rules,
}


class Tree(object):

    def __init__(
        self,
        cat: Category,
        children: Union[List['Tree'], List[Token]],
        op_string: str,
        op_symbol: str,
        head_is_left: bool = True,
    ) -> None:
        assert len({type(child) for child in children}) == 1, \
            "children must contain elements of a unique type"
        assert not isinstance(children[0], Tree) or len(children) in (1, 2), \
            "a tree cannot contain more than two children"
        assert not isinstance(children[0], Token) or len(children) == 1, \
            "a leaf node cannot contain more than one token object"

        self.cat = cat
        self.children = children
        self.op_string = op_string
        self.op_symbol = op_symbol
        self.head_is_left = head_is_left

    @staticmethod
    def make_terminal(
        word: Union[str, Token],
        cat: Category,
        op_string: str = 'lex',
        op_symbol: str = '<lex>',
    ) -> 'Tree':

        if isinstance(word, Token):
            token = word
        else:
            token = Token(word=word)

        return Tree(cat, [token], op_string, op_symbol)

    @staticmethod
    def make_binary(
        cat: Category,
        left: 'Tree',
        right: 'Tree',
        op_string: str,
        op_symbol: str,
        head_is_left: bool = True,
    ) -> 'Tree':
        return Tree(cat, [left, right], op_string, op_symbol, head_is_left)

    @staticmethod
    def make_unary(
        cat: Category,
        child: 'Tree',
        op_string: str = 'lex',
        op_symbol: str = '<un>'
    ) -> 'Tree':
        return Tree(cat, [child], op_string, op_symbol)

    @staticmethod
    def of_nltk_tree(tree) -> 'Tree':

        def rec(node):
            cat = Category.parse(node.label())
            if isinstance(node[0], str):
                word = node[0]
                return Tree.make_terminal(word, cat)
            else:
                children = [rec(child) for child in node]
                if len(children) == 1:
                    return Tree.make_unary(cat, children[0])
                else:
                    assert len(children) == 2
                    left, right = children
                    rule = guess_combinator_by_triplet(
                        BINARY_RULES[get_global_language()],
                        cat, left.cat, right.cat
                    )
                    return Tree.make_binary(
                        cat, left, right, rule.op_string, rule.op_symbol, rule.head_is_left
                    )

        return rec(tree)

    def __len__(self):
        return len(self.leaves)

    @property
    def leaves(self) -> List['Tree']:

        def rec(node):
            if node.is_leaf:
                result.append(node)
            else:
                for child in node.children:
                    rec(child)

        result = []
        rec(self)
        return result

    @property
    def tokens(self) -> List[Token]:
        return [leaf.children[0] for leaf in self.leaves]

    @property
    def token(self) -> Token:
        assert self.is_leaf, "Tree.token must be called on leaf objects"
        return self.children[0]

    @property
    def child(self):
        assert self.is_unary, "This node is not unary node! Please use `Tree.children`"
        return self.left_child

    @property
    def left_child(self):
        assert not self.is_leaf, "This node is leaf and does not have any child!"
        return self.children[0]

    @property
    def right_child(self):
        assert not self.is_leaf, "This node is leaf and does not have any child!"
        assert not self.is_unary, "This node does not have right child!"
        return self.children[1]

    @property
    def is_leaf(self):
        return (
            self.is_unary and isinstance(self.children[0], Token)
        )

    @property
    def word(self, token_key='word'):
        return ' '.join(token[token_key] for token in self.tokens)

    @property
    def is_unary(self) -> bool:
        return len(self.children) == 1

    # def __str__(self):
    #     return auto_of(self)

    # def __repr__(self):
    #     return auto_of(self)

    def nltk_tree(self):
        from nltk.tree import Tree

        def rec(node):
            if node.is_leaf:
                cat = node.cat
                children = [node.word]
            else:
                cat = node.cat
                children = [rec(child) for child in node.children]
            return Tree(str(cat), children)

        return rec(self)


class ScoredTree(NamedTuple):
    tree: Tree
    score: float


class ParseResult(NamedTuple):
    sentence_index: int
    tree_index: int
    tree: Tree
    tokens: List[Token]
    score: float


def iter_parse_results(
    nbest_trees: List[List[ScoredTree]],
    tagged_doc: List[List[Token]]
) -> Iterator[ParseResult]:

    for sentence_index, (trees, tokens) in enumerate(zip(nbest_trees, tagged_doc), 1):
        for tree_index, (tree, log_prob) in enumerate(trees, 1):
            yield ParseResult(
                sentence_index,
                tree_index,
                tree,
                tokens,
                log_prob,
            )



from depccg.annotator import annotate_using_janome, annotate_using_spacy
from depccg.types import Token


def test_spacy():
    sentences = [
        ["The Penn Treebank has recently implemented a new syntactic annotation scheme,",
         "designed to highlight aspects of predicate-argument structure."],
        ["This paper discusses the implementation of crucial aspects of this new annotation scheme."],
    ]

    tokens = [
        [
            Token(**{'word': 'The', 'pos': 'DT', 'entity': 'B-ORG',
                  'lemma': 'the', 'chunk': 'XX'}),
            Token(**{'word': 'Penn', 'pos': 'NNP', 'entity': 'I-ORG',
                  'lemma': 'penn', 'chunk': 'XX'}),
            Token(**{'word': 'Treebank', 'pos': 'NNP', 'entity': 'I-ORG',
                  'lemma': 'treebank', 'chunk': 'XX'}),
            Token(**{'word': 'has', 'pos': 'VBZ', 'entity': 'O',
                  'lemma': 'have', 'chunk': 'XX'}),
            Token(**{'word': 'recently', 'pos': 'RB', 'entity': 'O',
                  'lemma': 'recently', 'chunk': 'XX'}),
            Token(**{'word': 'implemented', 'pos': 'VBN', 'entity': 'O',
                  'lemma': 'implement', 'chunk': 'XX'}),
            Token(**{'word': 'a', 'pos': 'DT', 'entity': 'O',
                  'lemma': 'a', 'chunk': 'XX'}),
            Token(**{'word': 'new', 'pos': 'JJ', 'entity': 'O',
                  'lemma': 'new', 'chunk': 'XX'}),
            Token(**{'word': 'syntactic', 'pos': 'JJ', 'entity': 'O',
                  'lemma': 'syntactic', 'chunk': 'XX'}),
            Token(**{'word': 'annotation', 'pos': 'NN',
                     'entity': 'O', 'lemma': 'annotation', 'chunk': 'XX'}),
            Token(**{'word': 'scheme', 'pos': 'NN', 'entity': 'O',
                  'lemma': 'scheme', 'chunk': 'XX'}),
            Token(**{'word': ',', 'pos': ',', 'entity': 'O',
                  'lemma': ',', 'chunk': 'XX'}),
            Token(**{'word': 'designed', 'pos': 'VBN', 'entity': 'O',
                  'lemma': 'design', 'chunk': 'XX'}),
            Token(**{'word': 'to', 'pos': 'TO', 'entity': 'O',
                  'lemma': 'to', 'chunk': 'XX'}),
            Token(**{'word': 'highlight', 'pos': 'VB', 'entity': 'O',
                  'lemma': 'highlight', 'chunk': 'XX'}),
            Token(**{'word': 'aspects', 'pos': 'NNS', 'entity': 'O',
                  'lemma': 'aspect', 'chunk': 'XX'}),
            Token(**{'word': 'of', 'pos': 'IN', 'entity': 'O',
                  'lemma': 'of', 'chunk': 'XX'}),
            Token(**{'word': 'predicate', 'pos': 'NN', 'entity': 'O',
                  'lemma': 'predicate', 'chunk': 'XX'}),
            Token(**{'word': '-', 'pos': 'HYPH', 'entity': 'O',
                  'lemma': '-', 'chunk': 'XX'}),
            Token(**{'word': 'argument', 'pos': 'NN', 'entity': 'O', 'lemma': 'argument',
                     'chunk': 'XX'}),
            Token(**{'word': 'structure', 'pos': 'NN', 'entity': 'O',
                  'lemma': 'structure', 'chunk': 'XX'}),
            Token(**{'word': '.', 'pos': '.', 'entity': 'O',
                  'lemma': '.', 'chunk': 'XX'})
        ],
        [
            Token(**{'word': 'This', 'pos': 'DT', 'entity': 'O',
                  'lemma': 'this', 'chunk': 'XX'}),
            Token(**{'word': 'paper', 'pos': 'NN', 'entity': 'O',
                  'lemma': 'paper', 'chunk': 'XX'}),
            Token(**{'word': 'discusses', 'pos': 'VBZ', 'entity': 'O',
                  'lemma': 'discuss', 'chunk': 'XX'}),
            Token(**{'word': 'the', 'pos': 'DT', 'entity': 'O',
                  'lemma': 'the', 'chunk': 'XX'}),
            Token(**{'word': 'implementation', 'pos': 'NN', 'entity': 'O',
                  'lemma': 'implementation', 'chunk': 'XX'}),
            Token(**{'word': 'of', 'pos': 'IN', 'entity': 'O',
                  'lemma': 'of', 'chunk': 'XX'}),
            Token(**{'word': 'crucial', 'pos': 'JJ', 'entity': 'O',
                  'lemma': 'crucial', 'chunk': 'XX'}),
            Token(**{'word': 'aspects', 'pos': 'NNS', 'entity': 'O',
                  'lemma': 'aspect', 'chunk': 'XX'}),
            Token(**{'word': 'of', 'pos': 'IN', 'entity': 'O',
                  'lemma': 'of', 'chunk': 'XX'}),
            Token(**{'word': 'this', 'pos': 'DT', 'entity': 'O',
                  'lemma': 'this', 'chunk': 'XX'}),
            Token(**{'word': 'new', 'pos': 'JJ', 'entity': 'O',
                  'lemma': 'new', 'chunk': 'XX'}),
            Token(**{'word': 'annotation', 'pos': 'NN', 'entity': 'O',
                  'lemma': 'annotation', 'chunk': 'XX'}),
            Token(**{'word': 'scheme', 'pos': 'NN', 'entity': 'O',
                  'lemma': 'scheme', 'chunk': 'XX'}),
            Token(**{'word': '.', 'pos': '.', 'entity': 'O',
                  'lemma': '.', 'chunk': 'XX'})
        ]
    ]

    # raw_sentences = [
    #     ['The', 'Penn', 'Treebank', 'has', 'recently', 'implemented', 'a', 'new', 'syntactic', 'annotation', 'scheme',
    #         ',', 'designed', 'to', 'highlight', 'aspects', 'of', 'predicate', '-', 'argument', 'structure', '.'],
    #     ['This', 'paper', 'discusses', 'the', 'implementation', 'of', 'crucial',
    #         'aspects', 'of', 'this', 'new', 'annotation', 'scheme', '.']
    # ]

    assert tokens == annotate_using_spacy(sentences, tokenize=True)


test_spacy()


def test_janome():
    sentences = [
        ["メロスは激怒した。"],
        ["メロスには政治がわからぬ。"],
    ]

    tokens = [
        [
            Token(**{'word': 'メロス', 'surf': 'メロス', 'pos': '名詞', 'pos1': '一般', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '*', 'inflectionType': '*', 'reading': '*', 'base': 'メロス'}),
            Token(**{'word': 'は', 'surf': 'は', 'pos': '助詞', 'pos1': '係助詞', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '*', 'inflectionType': '*', 'reading': 'ハ', 'base': 'は'}),
            Token(**{'word': '激怒', 'surf': '激怒', 'pos': '名詞', 'pos1': 'サ変接続', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '*', 'inflectionType': '*', 'reading': 'ゲキド', 'base': '激怒'}),
            Token(**{'word': 'し', 'surf': 'し', 'pos': '動詞', 'pos1': '自立', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '連用形', 'inflectionType': 'サ変・スル', 'reading': 'シ', 'base': 'する'}),
            Token(**{'word': 'た', 'surf': 'た', 'pos': '助動詞', 'pos1': '*', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '基本形', 'inflectionType': '特殊・タ', 'reading': 'タ', 'base': 'た'}),
            Token(**{'word': '。', 'surf': '。', 'pos': '記号', 'pos1': '句点', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '*', 'inflectionType': '*', 'reading': '。', 'base': '。'})
        ],
        [
            Token(**{'word': 'メロス', 'surf': 'メロス', 'pos': '名詞', 'pos1': '一般', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '*', 'inflectionType': '*', 'reading': '*', 'base': 'メロス'}),
            Token(**{'word': 'に', 'surf': 'に', 'pos': '助詞', 'pos1': '格助詞', 'pos2': '一般', 'pos3': '*',
                     'inflectionForm': '*', 'inflectionType': '*', 'reading': 'ニ', 'base': 'に'}),
            Token(**{'word': 'は', 'surf': 'は', 'pos': '助詞', 'pos1': '係助詞', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '*', 'inflectionType': '*', 'reading': 'ハ', 'base': 'は'}),
            Token(**{'word': '政治', 'surf': '政治', 'pos': '名詞', 'pos1': '一般', 'pos2': '*', 'pos3': '*',
                     'inflectionForm': '*', 'inflectionType': '*', 'reading': 'セイジ', 'base': '政治'}),
            Token(**{'word': 'が', 'surf': 'が', 'pos': '助詞', 'pos1': '格助詞', 'pos2': '一般', 'pos3': '*',
                  'inflectionForm': '*', 'inflectionType': '*', 'reading': 'ガ', 'base': 'が'}),
            Token(**{'word': 'わから', 'surf': 'わから', 'pos': '動詞', 'pos1': '自立', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '未然形', 'inflectionType': '五段・ラ行', 'reading': 'ワカラ', 'base': 'わかる'}),
            Token(**{'word': 'ぬ', 'surf': 'ぬ', 'pos': '助動詞', 'pos1': '*', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '基本形', 'inflectionType': '特殊・ヌ', 'reading': 'ヌ', 'base': 'ぬ'}),
            Token(**{'word': '。', 'surf': '。', 'pos': '記号', 'pos1': '句点', 'pos2': '*', 'pos3': '*',
                  'inflectionForm': '*', 'inflectionType': '*', 'reading': '。', 'base': '。'})
        ]
    ]

    # raw_sentences = [
    #     ['メロス', 'は', '激怒', 'し', 'た', '。'],
    #     ['メロス', 'に', 'は', '政治', 'が', 'わから', 'ぬ', '。']
    # ]

    assert tokens == annotate_using_janome(sentences, tokenize=True)



# -*- coding: utf-8 -*-
#
#  Copyright 2015 Pascual Martinez-Gomez
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from nltk import FeatStruct
import re

class Category(object):
    """ Implements a CCG syntactic category with features. """

    def __init__(self, category):
        if isinstance(category, self.__class__):
            self.types = category.types
            self.type_features = category.type_features
        else:
            self.types = remove_feats_from_category(category)
            self.type_features = get_feats_from_category(category)

    def __repr__(self):
        return "Types: {0}\tFeats: {1}".format(self.types, self.type_features)

    def match(self, other):
        if not isinstance(other, self.__class__):
            return False
        if len(self.type_features) != len(other.type_features):
            return False
        # TODO: Ideally these substitutions should be precomputed for speed-up.
        types = re.sub(r'\\', r'\\\\', self.types)
        types = types.replace('|', '[/\\\]')
        types = types.replace('(', '\\(').replace(')', '\\)')
        if not re.fullmatch(types, other.types):
            return False
        return all([a.subsumes(b)
                    for (a, b) in zip(self.type_features, other.type_features)])

    def match_(self, other):
        return isinstance(other, self.__class__) \
          and len(self.type_features) == len(other.type_features) \
          and re.fullmatch(re.sub(r'\\', r'\\\\', self.types), other.types) \
          and all([a.subsumes(b) \
                   for (a, b) in zip(self.type_features, other.type_features)])

    def get_num_args(self):
        return len(self.type_features) - 1

def get_feats_from_category(category):
    r""" Returns the features of the syntactic category.
    category="S[mod=nm,form=base]" --> feats=['[mod=nm,form=base]']
    category="(S/S)\NP[mod=nm,case=nc]" --> feats=['', '', '[mod=nm,case=nc]']
    category="S[mod=nm,form=base]\NP[mod=nm,case=nc]" -->
      feats=['[mod=nm,form=base]', '[mod=nm,case=nc]']
    """
    feature_strings = re.findall(r'\w+(\[.+?\])*', category)
    features = []
    for feature_str in feature_strings:
        if feature_str == '':
            feature = FeatStruct({})
        else:
            attribute_value_dict = {}
            for attribute_value in feature_str.strip('[]').split(','):
                [attribute, value] = attribute_value.split('=')
                attribute_value_dict[attribute] = value
            feature = FeatStruct(attribute_value_dict)
        features.append(feature)
    return features

def remove_feats_from_category(category):
    r""" Remove features from syntactic category and returns only the category.
    category="S[mod=nm,form=base]" --> category="S"
    category="(S/S)\NP[mod=nm,case=nc]" --> category="(S/S)\NP"
    category="S[mod=nm,form=base]\NP[mod=nm,case=nc]" --> category="S\NP"
    """
    return re.sub(r'\[.+?\]', '', category)



import json
import logging
import re

from collections import defaultdict
from typing import NamedTuple, Union, List
from pathlib import Path


logging.basicConfig(
    format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
    level=logging.INFO
)
logger = logging.getLogger(__name__)


test = """
(TOP (Sm."L" (Sm."L" (CPt."L" (IP-MS."L" (IP-MS."L" (IP-MS."L" (IP-MS *pro*) (<IP-MS\IP-MS> 「)) (<IP-MS\IP-MS>."R" (<<IP-MS\IP-MS>/<IP-MS\IP-MS>>."L" (<<IP-MS\IP-MS>/<IP-MS\IP-MS>>."L" (<<IP-MS\IP-MS>/<IP-MS\IP-MS>>."L" (PPs *pro*) (<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>."R" (<<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>/<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>>."L" (<<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>/<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>> さあ) (<<<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>/<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>>\<<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>/<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>>> 、)) (<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>."FCLeft1" (<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>>."FCLeft1" (<PPs\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>> 流れ) (<<<IP-MS\IP-MS>/<IP-MS\IP-MS>>\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>> て)) (<<<IP-MS\IP-MS>/<IP-MS\IP-MS>>\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>> 来る)))) (<<<IP-MS\IP-MS>/<IP-MS\IP-MS>>\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>> ぞ)) (<<<IP-MS\IP-MS>/<IP-MS\IP-MS>>\<<IP-MS\IP-MS>/<IP-MS\IP-MS>>> 。)) (<IP-MS\IP-MS>."FCLeft1" (<IP-MS\IP-MS>."L" (PPs みんな) (<PPs\<IP-MS\IP-MS>> とれ)) (<IP-MS\IP-MS> 。)))) (<IP-MS\IP-MS> 」)) (<IP-MS\CPt> と)) (<CPt\Sm>."L" (PPs."L" (NP 一郎) (<NP\PPs> が)) (<PPs\<CPt\Sm>>."FCLeft2" (<PPs\<CPt\Sm>>."FCLeft2" (<PPs\<CPt\Sm>> 言い) (<Sm\Sm> まし)) (<Sm\Sm> た)))) (<Sm\Sm> 。)) (ID 771_aozora_Miyazawa-1934;JP))
"""


Category = Union['Atomic', 'Functor']
Node = Union['Leaf', 'Tree']


class Atomic(NamedTuple):
    string: str

    def __str__(self):
        return self.string


class Functor(NamedTuple):
    left: str
    slash: str
    right: str

    def __str__(self):
        """
        convert to CCG version
        """
        def rec(cat):
            if isinstance(cat, Atomic):
                return str(cat)
            right = rec(cat.right if cat.slash == '\\' else cat.left)
            left = rec(cat.left if cat.slash == '\\' else cat.right)
            return f'({right}{cat.slash}{left})'
        return rec(self)[1:-1]


class Leaf(NamedTuple):
    cat: str
    word: str
    pos: int

    def __str__(self):
        return f'({self.cat}, {self.word})'


class Tree(NamedTuple):
    cat: Category
    children: List[Node]

    def __str__(self):
        children = ', '.join(str(child) for child in self.children)
        return f'({self.cat}, ({children}))'


def drop_brackets(cat):
    if cat.startswith('<') and \
            cat.endswith('>') and \
            find_closing_bracket(cat) == len(cat) - 1:
        return cat[1:-1]
    else:
        return cat


def find_closing_bracket(source):
    open_brackets = 0
    for i, char in enumerate(source):
        open_brackets += char == '<'
        open_brackets -= char == '>'
        if open_brackets == 0:
            return i
    raise Exception("Mismatched brackets in string: " + source)


def find_non_nested_char(haystack, needles):
    open_brackets = 0
    for i in range(len(haystack) - 1, -1, -1):
        char = haystack[i]
        open_brackets += char == '<'
        open_brackets -= char == '>'
        if open_brackets == 0 and char in needles:
            return i
    return -1


FEATURE_PATTERN = re.compile(r'([a-z][a-z1-9]*)')


def parse_cat(cat):
    """
    ex. PPo1\<PPs\Sm>.h --> PPo1\(PPs\Sm)
    """
    rule = cat.rfind('.')
    if rule >= 0:
        cat = cat[:rule]
    cat = drop_brackets(cat)
    op_idx = find_non_nested_char(cat, '/\\')
    if op_idx == -1:
        return Atomic(FEATURE_PATTERN.sub('[\\1]', cat))
    else:
        left = parse_cat(cat[:op_idx])
        slash = cat[op_idx:op_idx + 1]
        right = parse_cat(cat[op_idx + 1:])
        return Functor(left, slash, right)


class KeyakiParser(object):
    def __init__(self, line):
        self.index = 0
        self.line = line
        self.items = self.lex(line)

    def lex(self, line):
        def is_bracket(string):
            return string in ['(', ')']

        flag = False
        res = []
        pos = 0

        line = line.replace('(', '( ').replace(')', ' )').split()
        for a, b in zip(line, line[1:]):
            if flag:
                flag = False
                continue
            elif not is_bracket(a) and not is_bracket(b):
                res.append(Leaf(parse_cat(a), b, pos))
                flag = True
                pos += 1
            else:
                res.append(a)
                flag = False
        res.append(')')
        return res

    def next(self):
        res = self.items[self.index]
        self.index += 1
        return res

    def peek(self):
        return self.items[self.index]

    def peek_next(self):
        return self.items[self.index + 1]

    def peek_prev(self):
        return self.items[self.index - 1]

    def parse(self):
        """
        parse (IP-MAT (INTJ ..) (INTJ ..))
        """
        if isinstance(self.peek_next(), Leaf):
            return self.parse_terminal()
        else:
            return self.parse_nonterminal()

    def parse_terminal(self):
        """
        parse ( INTJ えーっと )
        """
        assert self.next() == '('
        res = self.next()
        assert self.next() == ')'
        return res

    def parse_nonterminal(self):
        """
        parse ( IP-MAT ( INTJ .. ) ( INTJ .. ) )
        """
        assert self.next() == '(', self.index
        tag = parse_cat(self.next())
        children = []
        while self.peek() != ')':
            children.append(self.parse())
        assert self.next() == ')'
        res = Tree(tag, children)
        return res


def tree_is_to_be_used(tree):
    """
    discard the input tree if:
    1) the root is FRAG
    2) it contains empty category such as *T*
    3) a node with more than two children
    """
    def rec(node):
        if isinstance(node, Tree):
            if len(node.children) > 2:
                # logger.warn(f'more than 2 children: {tree}')
                return False
            return all(rec(child) for child in node.children)
        else:
            return not (node.word.startswith('*') and node.word.endswith('*'))
    if str(tree.cat) == 'FRAG':
        return False
    else:
        return rec(tree)


def read_keyaki(keyakipath, remove_top=True):
    for line in open(keyakipath):
        try:
            tree = KeyakiParser(line.strip()).parse()
        except AssertionError:
            continue
        if isinstance(tree, Tree) and remove_top:
            tree = tree.children[0]
        if tree_is_to_be_used(tree):
            yield tree


def get_leaves(tree):
    res = []

    def rec(tree):
        for child in tree.children:
            if isinstance(child, Tree):
                rec(child)
            else:
                assert isinstance(child, Leaf)
                res.append(child)

    if isinstance(tree, Tree):
        rec(tree)
    else:
        res.append(tree)
    return res


UNK = "*UNKNOWN*"
START = "*START*"
END = "*END*"
IGNORE = -1


class TrainingDataCreator(object):
    """
    create train & validation data
    """

    def __init__(self, filepath, word_freq_cut, char_freq_cut, cat_freq_cut):
        self.filepath = filepath
        # those categories whose frequency < freq_cut are discarded.
        self.word_freq_cut = word_freq_cut
        self.char_freq_cut = char_freq_cut
        self.cat_freq_cut = cat_freq_cut
        self.seen_rules = defaultdict(int)  # seen binary rules
        self.unary_rules = defaultdict(int)  # seen unary rules
        self.cats = defaultdict(int, {
            START: cat_freq_cut,
            END: cat_freq_cut
        })  # all cats
        self.words = defaultdict(int, {
            UNK: word_freq_cut,
            START: word_freq_cut,
            END: word_freq_cut
        })
        self.chars = defaultdict(int, {
            UNK: char_freq_cut,
            START: char_freq_cut,
            END: char_freq_cut
        })
        self.samples = []
        self.sents = []

    def _traverse(self, tree):
        if isinstance(tree, Leaf):
            self.cats[str(tree.cat)] += 1
            word = tree.word
            self.words[word] += 1
            for char in word:
                self.chars[char] += 1
        else:
            children = tree.children
            if len(children) == 1:
                rule = str(tree.cat), str(children[0].cat)
                self.unary_rules[rule] += 1
                self._traverse(children[0])
            else:
                if len(children) == 2:  # TODO
                    rule = str(children[0].cat), str(children[1].cat)
                    self.seen_rules[rule] += 1
                    self._traverse(children[0])
                    self._traverse(children[1])

    @staticmethod
    def _write(dct, filename):
        with open(filename, 'w') as f:
            logger.info(f'writing to {f.name}')
            for key, value in dct.items():
                print(f'{key} # {str(value)}', file=f)

    def _get_dependencies(self, tree, sent_len):
        def rec(subtree):
            if isinstance(subtree, Tree):
                children = subtree.children
                if len(children) == 2:
                    head = rec(children[1])
                    dep = rec(children[0])
                    res[dep] = head
                else:
                    head = rec(children[0])
                return head
            else:
                return subtree.pos

        res = [-1 for _ in range(sent_len)]
        rec(tree)
        res = [i + 1 for i in res]
        assert len(list(filter(lambda i: i == 0, res))) == 1, (res, str(tree))
        return res

    def _to_conll(self, out):
        for sent, (cats, deps) in self.samples:
            words = sent.split(' ')
            for i, (word, cat, dep) in enumerate(zip(words, cats, deps), 1):
                print(f'{i}\t{word}\t{cat}\t{dep}', file=out)
            print('', file=out)

    def _create_samples(self, trees):
        for tree in trees:
            tokens = get_leaves(tree)
            words = [token.word for token in tokens]
            cats = [str(token.cat) for token in tokens]
            deps = self._get_dependencies(tree, len(tokens))
            sent = ' '.join(words)
            self.sents.append(sent)
            self.samples.append((sent, [cats, deps]))

    @staticmethod
    def create_traindata(args):
        self = TrainingDataCreator(args.PATH,
                                   args.word_freq_cut,
                                   args.char_freq_cut,
                                   args.cat_freq_cut)

        trees = [tree for tree in read_keyaki(self.filepath)]
        # trees = [] # TODO
        # for line in open(self.filepath):
        #     try:
        #         trees.append(KeyakiParser(line.strip()).parse())
        #     except Exception:
        #         continue

        for tree in trees:
            self._traverse(tree)
        self._create_samples(trees)

        cats = {k: v for k, v in self.cats.items() if v >= self.cat_freq_cut}
        self._write(cats, args.OUT / 'target.txt')

        words = {k: v for k, v in self.words.items() if v >=
                 self.word_freq_cut}
        self._write(words, args.OUT / 'words.txt')

        chars = {k: v for k, v in self.chars.items() if v >=
                 self.char_freq_cut}
        self._write(chars, args.OUT / 'chars.txt')

        seen_rules = {f'{c1} {c2}': v for (c1, c2), v in self.seen_rules.items()
                      if c1 in cats and c2 in cats}
        self._write(seen_rules, args.OUT / 'seen_rules.txt')

        unary_rules = {f'{c1} {c2}': v for (
            c1, c2), v in self.unary_rules.items()}
        self._write(unary_rules, args.OUT / 'unary_rules.txt')

        with open(args.OUT / 'traindata.json', 'w') as f:
            logger.info(f'writing to {f.name}')
            json.dump(self.samples, f)

        with open(args.OUT / 'trainsents.txt', 'w') as f:
            logger.info(f'writing to {f.name}')
            for sent in self.sents:
                print(sent, file=f)

        with open(args.OUT / 'trainsents.conll', 'w') as f:
            logger.info(f'writing to {f.name}')
            self._to_conll(f)

    @staticmethod
    def create_testdata(args):
        self = TrainingDataCreator(args.PATH,
                                   args.word_freq_cut,
                                   args.char_freq_cut,
                                   args.cat_freq_cut)

        trees = [tree for tree in read_keyaki(self.filepath)]
        # trees = [] # TODO
        # for line in open(self.filepath):
        #     try:
        #         trees.append(KeyakiParser(line.strip()).parse())
        #     except Exception:
        #         continue

        self._create_samples(trees)
        with open(args.OUT / 'testdata.json', 'w') as f:
            logger.info(f'writing to {f.name}')
            json.dump(self.samples, f)

        with open(args.OUT / 'testsents.txt', 'w') as f:
            logger.info(f'writing to {f.name}')
            for sent in self.sents:
                print(sent, file=f)

        with open(args.OUT / 'testsents.conll', 'w') as f:
            logger.info(f'writing to {f.name}')
            self._to_conll(f)

    @staticmethod
    def convert_json(autopath):
        self = TrainingDataCreator(autopath, None, None, None)
        trees = [tree for tree in read_keyaki(self.filepath)]
        logger.info(f'loaded {len(trees)} trees')
        self._create_samples(trees)
        return self.samples


def convert_keyaki_to_json(keyakipath):
    return TrainingDataCreator.convert_json(keyakipath)


def main():
    import argparse
    parser = argparse.ArgumentParser('keyaki')

    parser.add_argument('PATH',
                        type=Path,
                        help='path to ccgbank data file')
    parser.add_argument('OUT',
                        type=Path,
                        help='output directory path')
    parser.add_argument('--cat-freq-cut',
                        type=int,
                        default=10,
                        help='only allow categories which appear >= freq-cut')
    parser.add_argument('--word-freq-cut',
                        type=int,
                        default=5,
                        help='only allow words which appear >= freq-cut')
    parser.add_argument('--afix-freq-cut',
                        type=int,
                        default=5,
                        help='only allow afixes which appear >= freq-cut')
    parser.add_argument('--char-freq-cut',
                        type=int,
                        default=5,
                        help='only allow characters which appear >= freq-cut')
    parser.add_argument('--mode',
                        choices=['train', 'test'],
                        default='train')

    args = parser.parse_args()

    if args.mode == 'train':
        TrainingDataCreator.create_traindata(args)
    else:
        TrainingDataCreator.create_testdata(args)


if __name__ == '__main__':
    main()



from typing import Iterator, List, Tuple
import re

from depccg.cat import Category
from depccg.tree import Tree
from depccg.types import Token
from depccg.tools.reader import ReaderResult


combinators = {
    'SSEQ', '>', '<', '>B', '<B1', '<B2', '<B3',
    '<B4', '>Bx1', '>Bx2', '>Bx3',
    'ADNext', 'ADNint', 'ADV0', 'ADV1', 'ADV2'
}

DEPENDENCY = re.compile(r'{.+?}')


def read_ccgbank(filepath: str) -> Iterator[ReaderResult]:
    """read Japanase CCGBank file.

    Args:
        filename (str): file name string

    Yields:
        Iterator[ReaderResult]: iterator object containing parse results
    """

    for i, line in enumerate(open(filepath)):
        line = line.strip()
        if len(line) == 0:
            continue
        tree, tokens = _JaCCGLineReader(line).parse()
        yield ReaderResult(str(i), tokens, tree)


class _JaCCGLineReader(object):
    def __init__(self, line: str) -> None:
        self.line = line
        self.index = 0
        self.word_id = -1
        self.tokens = []

    def next(self, target: str) -> str:
        end = self.line.find(target, self.index)
        result = self.line[self.index:end]
        self.index = end + 1
        return result

    def check(self, text: str, offset: int = 0) -> None:
        if self.line[self.index + offset] != text:
            raise RuntimeError('AutoLineReader.check catches parse error')

    def peek(self) -> str:
        return self.line[self.index]

    def parse(self) -> Tuple[Tree, List[Token]]:
        result = self.next_node()
        return result, self.tokens

    @property
    def next_node(self):
        end = self.line.find(' ', self.index)
        if self.line[self.index + 1:end] in combinators:
            return self.parse_tree
        else:
            return self.parse_leaf

    def parse_leaf(self) -> Tree:
        self.word_id += 1
        self.check('{')
        cat = self.next(' ')[1:]
        cat = cat[:cat.find('_')]
        cat = DEPENDENCY.sub('', cat)
        cat = Category.parse(cat)
        surf, base, pos1, pos2 = self.next('}')[:-1].split('/')
        token = Token(surf=surf, base=base, pos1=pos1, pos2=pos2)
        self.tokens.append(token)
        return Tree.make_terminal(surf, cat)

    def parse_tree(self) -> Tree:
        self.check('{')
        op_string = self.next(' ')
        cat = DEPENDENCY.sub('', self.next(' '))
        cat = Category.parse(cat)
        self.check('{')

        children = []
        while self.peek() != '}':
            children.append(self.next_node())
            if self.peek() == ' ':
                self.next(' ')

        self.next('}')

        if len(children) == 1:
            return Tree.make_unary(cat, children[0], op_string, op_string)
        else:
            assert len(
                children) == 2, f'failed to parse, invalid number of children: {self.line}'
            left, right = children
            return Tree.make_binary(cat, left, right, op_string, op_string)



from typing import Tuple, List
import numpy
import logging
from itertools import islice

from depccg.allennlp.predictor.supertagger_predictor import SupertaggerPredictor
from depccg.allennlp.dataset.ja_supertagging_dataset import JaSupertaggingDatasetReader
from depccg.allennlp.dataset.supertagging_dataset import TritrainSupertaggingDatasetReader
from depccg.allennlp.dataset.supertagging_dataset import SupertaggingDatasetReader
from depccg.allennlp.models.supertagger import Supertagger
from allennlp.models.archival import load_archive

from depccg.types import ScoringResult

logger = logging.getLogger(__name__)


def lazy_groups_of(iterator, group_size):
    return iter(lambda: list(islice(iterator, 0, group_size)), [])


class AllennlpSupertagger(object):
    def __init__(self, predictor):
        self.predictor = predictor
        self.dataset_reader = predictor._dataset_reader

    def predict_doc(
        self,
        splitted,
        batchsize=32,
    ) -> Tuple[List[ScoringResult], List[str]]:

        instances = (
            self.dataset_reader.text_to_instance(' '.join(sentence))
            for sentence in splitted
        )

        categories = None
        scores = []
        for batch in lazy_groups_of(instances, batchsize):
            for json_dict in self.predictor.predict_batch_instance(batch):
                if categories is None:
                    categories = list(json_dict['categories'])
                dep_scores = numpy.array(json_dict['heads']) \
                    .reshape(json_dict['heads_shape']) \
                    .astype(numpy.float32)
                tag_scores = numpy.array(json_dict['head_tags']) \
                    .reshape(json_dict['head_tags_shape']) \
                    .astype(numpy.float32)

                scores.append(ScoringResult(tag_scores, dep_scores))
        return scores, categories


def load_allennlp_tagger(
    model_path: str,
    device: int = -1
) -> AllennlpSupertagger:

    if device >= 0:
        logger.info(f'sending the supertagger to gpu: {device}')
    archive = load_archive(model_path, cuda_device=device)
    predictor = SupertaggerPredictor.from_archive(
        archive, 'supertagger-predictor'
    )
    return AllennlpSupertagger(predictor)




import json
from depccg.tools.reader import read_auto
from depccg import utils
from collections import defaultdict
from pathlib import Path
import logging


logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
                    level=logging.INFO)
logger = logging.getLogger(__name__)


UNK = "*UNKNOWN*"
OOR2 = "OOR2"
OOR3 = "OOR3"
OOR4 = "OOR4"
START = "*START*"
END = "*END*"
IGNORE = -1
MISS = -2


def get_suffix(word):
    return [word[-1],
            word[-2:] if len(word) > 1 else OOR2,
            word[-3:] if len(word) > 2 else OOR3,
            word[-4:] if len(word) > 3 else OOR4]


def get_prefix(word):
    return [word[0],
            word[:2] if len(word) > 1 else OOR2,
            word[:3] if len(word) > 2 else OOR3,
            word[:4] if len(word) > 3 else OOR4]


class TrainingDataCreator(object):
    def __init__(self, filepath, word_freq_cut, cat_freq_cut, afix_freq_cut):
        self.filepath = filepath
        # those categories whose frequency < freq_cut are discarded.
        self.word_freq_cut = word_freq_cut
        self.cat_freq_cut = cat_freq_cut
        self.afix_freq_cut = afix_freq_cut
        self.seen_rules = defaultdict(int)  # seen binary rules
        self.unary_rules = defaultdict(int)  # seen unary rules
        self.cats = defaultdict(int)  # all cats
        self.words = defaultdict(int, {
            UNK: word_freq_cut,
            START: word_freq_cut,
            END: word_freq_cut
        })
        afix_defaults = {key: afix_freq_cut for key in [
            UNK, START, END, OOR2, OOR3, OOR4]}
        self.prefixes = defaultdict(int, afix_defaults)
        self.suffixes = defaultdict(int, afix_defaults)
        self.samples = []
        self.sents = []

    def _traverse(self, tree):
        if tree.is_leaf:
            self.cats[str(tree.cat)] += 1
            word = utils.normalize(tree.word)
            self.words[word.lower()] += 1
            for f in get_suffix(word):
                self.suffixes[f] += 1
            for f in get_prefix(word):
                self.prefixes[f] += 1
        else:
            children = tree.children
            if len(children) == 1:
                rule = str(tree.cat), str(children[0].cat)
                self.unary_rules[rule] += 1
                self._traverse(children[0])
            else:
                rule = str(children[0].cat), str(children[1].cat)
                self.seen_rules[rule] += 1
                self._traverse(children[0])
                self._traverse(children[1])

    @staticmethod
    def _write(dct, filename):
        with open(filename, 'w') as f:
            logger.info(f'writing to {f.name}')
            for key, value in dct.items():
                print(f'{key} # {str(value)}', file=f)

    def _get_dependencies(self, tree, sent_len):
        counter = 0

        def rec(subtree):
            nonlocal counter
            if not subtree.is_leaf:
                children = subtree.children
                if len(children) == 2:
                    head = rec(children[0])
                    dep = rec(children[1])
                    res[dep] = head
                else:
                    head = rec(children[0])
                return head
            else:
                head = counter
                counter += 1
                return head

        res = [-1 for _ in range(sent_len)]
        rec(tree)
        res = [i + 1 for i in res]
        assert len(list(filter(lambda i: i == 0, res))) == 1
        return res

    def _to_conll(self, out):
        for sent, [cats, deps] in self.samples:
            sent = sent.split(' ')
            for i, (word, cat, dep) in enumerate(zip(sent, cats, deps), 1):
                print(
                    f'{i}\t{word}\t{word}\tPOS\tPOS\t_\t{dep}\tnone\t_\t{cat}', file=out)
            print('', file=out)

    def _create_samples(self, trees):
        for tree in trees:
            tokens = tree.leaves
            words = [utils.normalize(token.word) for token in tokens]
            cats = [str(token.cat) for token in tokens]
            deps = self._get_dependencies(tree, len(tokens))
            sent = ' '.join(words)
            self.sents.append(sent)
            self.samples.append((sent, [cats, deps]))

    @staticmethod
    def create_traindata(args):
        self = TrainingDataCreator(args.PATH,
                                   args.word_freq_cut,
                                   args.cat_freq_cut,
                                   args.afix_freq_cut)

        trees = [tree for _, _, tree in read_auto(
            self.filepath) if tree.word != 'FAILED']
        logger.info(f'loaded {len(trees)} trees')
        for tree in trees:
            self._traverse(tree)
        self._create_samples(trees)

        cats = {k: v for k, v in self.cats.items() if v >= self.cat_freq_cut}
        self._write(cats, args.OUT / 'target.txt')

        words = {k: v for k, v in self.words.items() if v >=
                 self.word_freq_cut}
        self._write(words, args.OUT / 'words.txt')

        suffixes = {k: v for k, v in self.suffixes.items() if v >=
                    self.afix_freq_cut}
        self._write(suffixes, args.OUT / 'suffixes.txt')

        prefixes = {k: v for k, v in self.prefixes.items() if v >=
                    self.afix_freq_cut}
        self._write(prefixes, args.OUT / 'prefixes.txt')

        seen_rules = {f'{c1} {c2}': v for (c1, c2), v in self.seen_rules.items()
                      if c1 in cats and c2 in cats}
        self._write(seen_rules, args.OUT / 'seen_rules.txt')

        unary_rules = {f'{c1} {c2}': v for (
            c1, c2), v in self.unary_rules.items()}

        self._write(unary_rules, args.OUT / 'unary_rules.txt')

        with open(args.OUT / 'traindata.json', 'w') as f:
            logger.info(f'writing to {f.name}')
            json.dump(self.samples, f)

        with open(args.OUT / 'trainsents.txt', 'w') as f:
            logger.info(f'writing to {f.name}')
            for sent in self.sents:
                print(sent, file=f)

        with open(args.OUT / 'trainsents.conll', 'w') as f:
            logger.info(f'writing to {f.name}')
            self._to_conll(f)

    @staticmethod
    def create_testdata(args):
        self = TrainingDataCreator(args.PATH,
                                   args.word_freq_cut,
                                   args.cat_freq_cut,
                                   args.afix_freq_cut)

        trees = [tree for _, _, tree in read_auto(self.filepath)]
        self._create_samples(trees)

        with open(args.OUT / 'testdata.json', 'w') as f:
            logger.info(f'writing to {f.name}')
            json.dump(self.samples, f)

        with open(args.OUT / 'testsents.txt', 'w') as f:
            logger.info(f'writing to {f.name}')
            for sent in self.sents:
                print(sent, file=f)

        with open(args.OUT / 'testsents.conll', 'w') as f:
            logger.info(f'writing to {f.name}')
            self._to_conll(f)

    @staticmethod
    def convert_json(autopath):
        self = TrainingDataCreator(autopath, None, None, None)
        trees = [tree for _, _, tree in read_auto(
            self.filepath) if tree.word != 'FAILED']
        logger.info(f'loaded {len(trees)} trees')
        self._create_samples(trees)
        return self.samples


def convert_auto_to_json(autopath):
    return TrainingDataCreator.convert_json(autopath)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()

    # Creating training data
    parser.add_argument('PATH',
                        type=Path,
                        help='path to conll data file')
    parser.add_argument('OUT',
                        type=Path,
                        help='output directory path')
    parser.add_argument('--cat-freq-cut',
                        type=int,
                        default=10,
                        help='only allow categories which appear >= freq-cut')
    parser.add_argument('--word-freq-cut',
                        type=int,
                        default=5,
                        help='only allow words which appear >= freq-cut')
    parser.add_argument('--afix-freq-cut',
                        type=int,
                        default=5,
                        help='only allow afixes which appear >= freq-cut')
    parser.add_argument('--mode',
                        choices=['train', 'test'],
                        default='train')

    args = parser.parse_args()
    if args.mode == 'train':
        TrainingDataCreator.create_traindata(args)
    else:
        TrainingDataCreator.create_testdata(args)



from typing import Optional, List, TypeVar, Tuple, Set, Dict
from string import ascii_letters
from depccg.cat import Category
from depccg.unification import Unification
from depccg.types import Combinator, CombinatorResult

X = TypeVar('X')
Pair = Tuple[X, X]


def _match(x: Category, y: Category) -> bool:
    if x.is_functor and y.is_functor:
        return _match(x.left, y.left) and _match(x.right, y.right)
    elif x.is_atomic and y.is_atomic:
        return (
            x.base == y.base
            and x.feature.unifies(y.feature)
            and y.feature.unifies(x.feature)
        )
    return False


def _is_modifier(x: Category) -> bool:
    return x.is_functor and x.left == x.right


def _is_punct(x: Category) -> bool:
    if x.is_functor:
        return False

    return (
        not x.base[0] in ascii_letters
        or x.base in ("LRB", "RRB", "LQU", "RQU")
    )


def _is_type_raised(x: Category) -> bool:
    if x.is_atomic:
        return False
    return (
        x.right.is_functor and x.right.left == x.left
    )


def forward_application(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("a/b", "b")
    if uni(x, y):
        result = y if _is_modifier(x) else uni['a']
        return CombinatorResult(
            cat=result,
            op_string="fa",
            op_symbol=">",
            head_is_left=True,
        )
    return None


def backward_application(x: Category, y: Category) -> Optional[CombinatorResult]:
    if x == 'S[dcl]' and y == 'S[em]\\S[em]':
        result = x
    else:
        uni = Unification("b", "a\\b")
        if uni(x, y):
            result = x if _is_modifier(y) else uni['a']
        else:
            return None

    return CombinatorResult(
        cat=result,
        op_string="ba",
        op_symbol="<",
        head_is_left=True,
    )


def forward_composition(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("a/b", "b/c")
    if uni(x, y):
        result = y if _is_modifier(x) else uni['a'] / uni['c']
        return CombinatorResult(
            cat=result,
            op_string="fc",
            op_symbol=">B",
            head_is_left=True,
        )
    return None


def backward_composition(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("b/c", "a\\b")
    if uni(x, y):
        if str(uni["b"]) in ("N", "NP"):
            return None
        result = x if _is_modifier(y) else uni['a'] / uni['c']
        return CombinatorResult(
            cat=result,
            op_string="bx",
            op_symbol="<B",
            head_is_left=True,
        )
    return None


def generalized_forward_composition(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("a/b", "(b/c)|d")
    if uni(x, y):
        result = y if _is_modifier(x) else y.functor(
            (uni['a'] / uni['c']), uni['d'])
        return CombinatorResult(
            cat=result,
            op_string="gfc",
            op_symbol=">B",
            head_is_left=True,
        )
    return None


def generalized_backward_composition(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("(b/c)|d", "a/b")
    if uni(x, y):
        if str(uni["b"]) in ("N", "NP"):
            return None
        result = x if _is_modifier(y) else x.functor(
            (uni['a'] / uni['c']), uni['d'])
        return CombinatorResult(
            cat=result,
            op_string="gbx",
            op_symbol="<B",
            head_is_left=True,
        )
    return None


def conjunction(x: Category, y: Category) -> Optional[CombinatorResult]:
    if (
        not _is_punct(y)
        and not _is_type_raised(y)
        and x in (",", ";", "conj")
        and not (y ^ "NP\\NP")
    ):
        result = y | y
        return CombinatorResult(
            cat=result,
            op_string="conj",
            op_symbol="<Φ>",
            head_is_left=True,
        )
    return None


def conjunction2(x: Category, y: Category) -> Optional[CombinatorResult]:
    if x == "conj" and y == "NP\\NP":
        result = y
        return CombinatorResult(
            cat=result,
            op_string="conj",
            op_symbol="<Φ>",
            head_is_left=True,
        )
    return None


def remove_punctuation1(x: Category, y: Category) -> Optional[CombinatorResult]:
    if _is_punct(x):
        result = y
        return CombinatorResult(
            cat=result,
            op_string="lp",
            op_symbol="<lp>",
            head_is_left=True,
        )
    return None


def remove_punctuation2(x: Category, y: Category) -> Optional[CombinatorResult]:
    if _is_punct(y):
        result = x
        return CombinatorResult(
            cat=result,
            op_string="rp",
            op_symbol="<rp>",
            head_is_left=True,
        )
    return None


def remove_punctuation_left(x: Category, y: Category) -> Optional[CombinatorResult]:
    if x in ("LQU", "LRB"):
        result = y | y
        return CombinatorResult(
            cat=result,
            op_string="lp",
            op_symbol="<lp>",
            head_is_left=True,
        )
    return None


def comma_vp_to_adv(x: Category, y: Category) -> Optional[CombinatorResult]:
    if x == "," and y in ("S[ng]\\NP", "S[pss]\\NP"):
        result = Category.parse("(S\\NP)\\(S\\NP)")
        return CombinatorResult(
            cat=result,
            op_string="lp",
            op_symbol="<*>",
            head_is_left=True
        )
    return None


def parenthetical_direct_speech(x: Category, y: Category) -> Optional[CombinatorResult]:
    if x == "," and y == "S[dcl]/S[dcl]":
        result = Category.parse("(S\\NP)/(S\\NP)")
        return CombinatorResult(
            cat=result,
            op_string="lp",
            op_symbol="<*>",
            head_is_left=True
        )
    return None


combinators: List[Combinator] = [
    forward_application,
    backward_application,
    forward_composition,
    backward_composition,
    generalized_forward_composition,
    generalized_backward_composition,
    conjunction,
    conjunction2,
    remove_punctuation1,
    remove_punctuation2,
    remove_punctuation_left,
    comma_vp_to_adv,
    parenthetical_direct_speech,
]


def apply_binary_rules(
    x: Category,
    y: Category,
    seen_rules: Optional[Set[Pair[Category]]] = None,
) -> List[CombinatorResult]:
    key = (x.clear_features('nb'), y.clear_features('nb'))
    seen_key = (
        x.clear_features('X', 'nb'), y.clear_features('X', 'nb')
    )
    results = []
    if seen_rules is None or seen_key in seen_rules:
        for combinator in combinators:
            result = combinator(*key)
            if result is not None:
                results.append(result)

    return results


def apply_unary_rules(
    x: Category,
    unary_rules: Dict[Category, List[Category]]
) -> List[CombinatorResult]:
    if x not in unary_rules:
        return []
    results = []
    for result in unary_rules[x]:
        type_raised = (
            x.is_atomic
            and x.base in ('NP', 'PP')
            and _is_type_raised(result)
        )

        results.append(
            CombinatorResult(
                cat=result,
                op_string='tr' if type_raised else 'lex',
                op_symbol='<un>',
                head_is_left=True,
            )
        )
    return results
